import enum as _enum


class CommonLabels(_enum.StrEnum):
    label = "label"
    label_unique = "label_unique"
    from_bus = "from"
    from_unique = "from_unique"
    to = "to"
    to_unique = "to_unique"
    connect = "connect"
    connect_unique = "connect_unique"
    building = "building"


class NodeKeys(_enum.StrEnum):
    links = 'links'
    buses = "buses"
    grid_connection = "grid_connection"
    commodity_sources = "commodity_sources"
    solar = "solar"
    transformers = "transformers"
    demand = "demand"
    storages = "storages"
    stratified_storage = "stratified_storage"
    ice_storage = "ice_storage"
    profiles = "profiles"


class NodeKeysOptional:
    # TODO: move other optional ones here and fix any arising issues.
    building_model_parameters = "building_model_parameters"


class CsvInputFilePathsRelative(_enum.StrEnum):
    buses = "buses.csv"
    grid_connection = "grid_connection.csv"
    commodity_sources = "commodity_sources.csv"
    solar = "solar.csv"
    transformers = "transformers.csv"
    demand = "demand.csv"
    storages = "storages.csv"
    stratified_storage = "stratified_storage.csv"
    ice_storage = "ice_storage.csv"
    profiles = "profiles.csv"
    links = 'links.csv'


class BusesLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    excess = "excess"
    excess_costs = "excess costs"
    active = "active"


class BusTypes(_enum.StrEnum):
    naturalGasBus = "naturalGasBus"
    gridBus = "gridBus"
    electricityBus = "electricityBus"
    electricityProdBus = "electricityProdBus"
    electricityInBus = "electricityInBus"
    shSourceBus = "shSourceBus"
    spaceHeatingBus = "spaceHeatingBus"
    shDemandBus = "shDemandBus"
    domesticHotWaterBus = "domesticHotWaterBus"
    dhwDemandBus = "dhwDemandBus"
    dhwStorageBus = "dhwStorageBus"
    solarConnectBus = "solarConnectBus"
    heatBus = "heatBus"
    heatDemandBus = "heatDemandBus"
    lowTSourceBus = "lowTSourceBus"
    lowTSinkBus = "lowTSinkBus"
    dhHeatInBus = "districtHeatingInBus"
    dhHeatBus = "districtHeatingBus"
    electricitySourceBusPVT = "elSource_pvt"


class CommoditySourcesLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    to = CommonLabels.to.value
    variable_costs = "variable costs"
    CO2_impact = "CO2 impact"
    active = "active"


class CommoditySourceTypes(_enum.StrEnum):
    naturalGasResource = "naturalGasResource"
    electricityResource = "electricityResource"


class DemandLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    active = "active"
    from_bus = CommonLabels.from_bus.value
    fixed = "fixed"
    nominal_value = "nominal value"
    building_model = "building model"


class DemandTypes(_enum.StrEnum):
    electricityDemand = "electricityDemand"
    spaceHeatingDemand = "spaceHeatingDemand"
    domesticHotWaterDemand = "domesticHotWaterDemand"


class GridConnectionLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    from_bus = CommonLabels.from_bus.value
    to = CommonLabels.to.value
    efficiency = "efficiency"


class GridConnectionTypes(_enum.StrEnum):
    gridElectricity = "gridElectricity"
    electricitySource = "electricitySource"
    producedElectricity = "producedElectricity"
    domesticHotWater = "domesticHotWater"
    shSource = "shSource"
    spaceHeating = "spaceHeating"


class LinksLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    active = "active"
    efficiency = "efficiency"
    invest_base = "invest_base"
    invest_cap = "invest_cap"
    investment = "investment"


class LinksTypes(_enum.StrEnum):
    electricityLink = "electricityLink"
    shLink = "shLink"
    dhwLink = "dhwLink"
    heatLink0 = "heatLink0"
    heatLink2 = "heatLink2"
    lowTempHeatLink = "lowTempHeatLink"
    dhLink = "dhLink"


class ProfileLabels(_enum.StrEnum):
    name = "name"
    path = "path"
    info = "INFO"


class ProfileTypes(_enum.StrEnum):
    demand = "demand_profiles"  # mandatory
    weather = "weather_data"  # mandatory
    # TODO: add missing profile types  # pylint: disable=fixme
    internal_gains = "internal_gains"  # not mandatory
    building_model_params = "building_model_params"  # not mandatory
    fixed_sources = "fixed_source_profiles" # not mandatory


class SolarLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    active = "active"
    from_bus = CommonLabels.from_bus.value
    to = CommonLabels.to.value
    connect = "connect"
    electrical_consumption = "electrical_consumption"
    peripheral_losses = "peripheral_losses"
    latitude = "latitude"
    longitude = "longitude"
    tilt = "tilt"
    azimuth = "azimuth"
    efficiency = "efficiency"
    eta_0 = "eta_0"
    a_1 = "a_1"
    a_2 = "a_2"
    temp_collector_inlet = "temp_collector_inlet"
    delta_temp_n = "delta_temp_n"
    capacity_max = "capacity_max"
    capacity_min = "capacity_min"
    lifetime = "lifetime"
    maintenance = "maintenance"
    installation = "installation"
    planification = "planification"
    invest_base = "invest_base"
    invest_cap = "invest_cap"
    heat_impact = "heat_impact"
    elec_impact = "elec_impact"
    impact_cap = "impact_cap"
    roof_area = "roof_area"
    zenith_angle = "zenith_angle"


class SolarTypes(_enum.StrEnum):
    solarCollector = "solarCollector"
    pv = "pv"


class StorageLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    active = "active"
    from_bus = CommonLabels.from_bus.value
    to = CommonLabels.to.value
    efficiency_inflow = "efficiency inflow"
    efficiency_outflow = "efficiency outflow"
    initial_capacity = "initial capacity"
    initial_temp = "initial_temp"
    capacity_min = "capacity min"
    capacity_max = "capacity max"
    capacity_loss = "capacity loss"
    lifetime = "lifetime"
    maintenance = "maintenance"
    installation = "installation"
    planification = "planification"
    invest_base = "invest_base"
    invest_cap = "invest_cap"
    heat_impact = "heat_impact"
    elec_impact = "elec_impact"
    impact_cap = "impact_cap"


class StorageTypes(_enum.StrEnum):
    electricalStorage = "electricalStorage"
    shStorage = "shStorage"
    dhwStorage = "dhwStorage"
    pitStorage = "pitStorage"
    tankGenericStorage = "tankGenericStorage"
    pitGenericStorage = "pitGenericStorage"
    boreholeGenericStorage = "boreholeGenericStorage"
    aquifierGenericStorage = "aquifierGenericStorage"
    thermalStorage = "thermalStorage"
    coolingBufferStorage = "coolingBufferStorage"


class StratifiedStorageLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    diameter = "diameter"
    temp_h = "temp_h"
    temp_c = "temp_c"
    temp_env = "temp_env"
    inflow_conversion_factor = "inflow_conversion_factor"
    outflow_conversion_factor = "outflow_conversion_factor"
    s_iso = "s_iso"
    lamb_iso = "lamb_iso"
    alpha_inside = "alpha_inside"
    alpha_outside = "alpha_outside"


class StratifiedStorageTypes(_enum.StrEnum):
    dhwStorage = "dhwStorage"
    shStorage = "shStorage"


class IceStorageLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    max_ice_fraction = "max_ice_fraction"
    rho_fluid = "rho_fluid"
    h_fluid = "h_fluid"
    cp_fluid = "cp_fluid"
    UA_tank = "UA_tank"
    inflow_conversion_factor = "inflow_conversion_factor"
    outflow_conversion_factor = "outflow_conversion_factor"


class IceStorageTypes(_enum.StrEnum):
    iceStorage = "iceStorage"


class TransformerLabels(_enum.StrEnum):
    label = CommonLabels.label.value
    building = "building"
    active = "active"
    from_bus = CommonLabels.from_bus.value
    to = CommonLabels.to.value
    efficiency = "efficiency"
    capacity_DHW = "capacity_DHW"
    capacity_SH = "capacity_SH"
    capacity_el = "capacity_el"
    capacity_min = "capacity_min"
    lifetime = "lifetime"
    maintenance = "maintenance"
    installation = "installation"
    planification = "planification"
    invest_base = "invest_base"
    invest_cap = "invest_cap"
    heat_impact = "heat_impact"
    elec_impact = "elec_impact"
    impact_cap = "impact_cap"

class HeatPumpCoefficientLabels(_enum.StrEnum):
    coef_W = "coef_W"
    coef_Q = "coef_Q"


class TransformerTypes(_enum.StrEnum):
    # TODO: currently the investment object is applied to the input flow for all transformer types
    # If this changes, we need to split this class into
    # (1) transformers with investment on inflow (this class would then be used while converting input capacities to output capacities)
    # (2) transformers with investment on outflow
    CHP = "CHP"
    HP = "HP"
    GWHP = "GWHP"
    GasBoiler = "GasBoiler"
    BiomassBoiler = "BiomassBoiler"
    OilBoiler = "OilBoiler"
    ElectricRod = "ElectricRod"
    Chiller = "Chiller"


class BuildingModelParameters(_enum.StrEnum):
    building_unique = "building_unique"
    Building_Number = "Building Number"
    Circuit = "Circuit"
    gAreaWindows = "gAreaWindows"
    rDistribution = "rDistribution"
    cDistribution = "cDistribution"
    rIndoor = "rIndoor"
    cIndoor = "cIndoor"
    rWall = "rWall"
    cWall = "cWall"
    qDistributionMin = "qDistributionMin"
    qDistributionMax = "qDistributionMax"
    tIndoorMin = "tIndoorMin"
    tIndoorMax = "tIndoorMax"
    tIndoorInit = "tIndoorInit"
    tDistributionInit = "tDistributionInit"
    tWallInit = "tWallInit"
    tIndoorDay = "tIndoorDay"
    tIndoorNight = "tIndoorNight"



import importlib as _ilib
import os as _os
import pathlib as _pl
import subprocess as _sp
import sys as _sys
import typing as _tp
import unittest as _ut

import matplotlib as _mpl
import pandas as _pd
from matplotlib import pyplot as _plt

import optihood as oh

ROOT_DIR = _pl.Path(oh.__file__).resolve().parents[1]
TEST_RESULTS_DIR = ROOT_DIR / "results"
ROOT_DATA_DIR = ROOT_DIR / "data"
EXAMPLE_SCRIPT_DIR = ROOT_DATA_DIR / "examples"
EXAMPLE_RESULTS_DIR = ROOT_DATA_DIR / "results"


def get_current_file_dir(file) -> _pl.Path:
    return _pl.Path(file).resolve().parent


def compare_txt_files(testCase: _ut.TestCase, file_path1: _pl.Path, file_path2: _pl.Path):
    with open(file_path1) as f1, open(file_path2) as f2:
        _ut.TestCase.assertListEqual(testCase, list(f1), list(f2))


def check_sheet_names(testCase: _ut.TestCase, data: _pd.ExcelFile, sheet_names_expected: _tp.Sequence[str]):
    """May need to have sheets ordered alphabetically."""
    testCase.assertListEqual(data.sheet_names, sheet_names_expected)


def plot_dfs_and_differences(df_new: _pd.DataFrame, df_expected: _pd.DataFrame, sheet_name: str):
    # =======
    # The following is needed to plot the differences as part of a pycharm test run.
    _mpl.use("QtAgg")
    # =======
    fig, axs = _plt.subplots(3, 1)
    df_new.plot(ax=axs[0])
    df_expected.plot(ax=axs[1])
    if df_new.shape == df_expected.shape:
        # avoids str - str errors.
        df_diff = df_new.select_dtypes(exclude=[object]) - df_expected.select_dtypes(exclude=[object])
        df_diff.plot(ax=axs[2])
    else:
        sheet_name += " SHAPE MISMATCH!"
    axs[0].set_title(f"sheet name: {sheet_name}")


def compare_xls_files(
        testCase: _ut.TestCase,
        file_path: _pl.Path,
        file_path_expected: _pl.Path,
        sheet_names_expected: _tp.Sequence[str],
        abs_tolerance: _tp.Optional[float] = None,
        rel_tolerance: _tp.Optional[float] = None,
        manual_test: bool = False,
):
    """Used to provide better feedback when comparing xls files.
    Assembles all errors before raising them.
    Manual testing allows plotting of the differences, given equal dataframes.

    When providing the abs_tolerance and rel_tolerance, the comparison uses the
    relative tolerance (`rtol`) and absolute tolerance (`atol`)
    parameters to determine if two values are considered close,
    according to the formula: `|a - b| <= (atol + rtol * |b|)
    """
    with _pd.ExcelFile(file_path) as data:
        with _pd.ExcelFile(file_path_expected) as expected_data:
            check_sheet_names(testCase, data, sheet_names_expected)
            errors = []

            print("")
            for sheet in sheet_names_expected:
                print(sheet)
                df_new = data.parse(sheet)
                df_expected = expected_data.parse(sheet)

                compare_dataframes(abs_tolerance, df_expected, df_new, errors, manual_test, rel_tolerance, sheet)
    if errors:
        if manual_test:
            _plt.show(block=True)
        raise ExceptionGroup(f"found {len(errors)} errors in {file_path}", errors)


def compare_csv_files(file_path: _pl.Path,
                      file_path_expected: _pl.Path,
                      name_for_df: str,
                      abs_tolerance: _tp.Optional[float] = None,
                      rel_tolerance: _tp.Optional[float] = None,
                      manual_test: bool = False,
                      ):
    df_new = _pd.read_csv(file_path)
    df_expected = _pd.read_csv(file_path_expected)

    errors = []
    compare_dataframes(abs_tolerance, df_expected, df_new, errors, manual_test, rel_tolerance, name_for_df)

    if errors:
        if manual_test:
            _plt.show(block=True)
        raise ExceptionGroup(f"found {len(errors)} errors in {file_path}", errors)


def compare_dataframes(abs_tolerance, df_expected, df_new, errors, manual_test, rel_tolerance, sheet):
    try:
        if abs_tolerance and rel_tolerance:
            _pd.testing.assert_frame_equal(df_new, df_expected, atol=abs_tolerance, rtol=rel_tolerance,
                                           check_exact=False)
        elif abs_tolerance:
            _pd.testing.assert_frame_equal(df_new, df_expected, atol=abs_tolerance, check_exact=False)
        elif rel_tolerance:
            _pd.testing.assert_frame_equal(df_new, df_expected, rtol=rel_tolerance, check_exact=False)
        else:
            _pd.testing.assert_frame_equal(df_new, df_expected, check_exact=True)
    except AssertionError as current_error:
        """Optihood doesn't export the results in a consistent way.
        Therefore, this hack reorders the results.
        Instead, the export should be ordered consistently.
        """

        df_new = df_new.sort_values(by=[df_new.columns[0]], ignore_index=True)
        df_expected = df_expected.sort_values(by=[df_new.columns[0]], ignore_index=True)

        """ The dType of a column sometimes gets set to int instead of float. 
            This reduces the feedback of the test to a dType check.
            Ignoring the dType ensures better feedback.
        """
        try:
            if abs_tolerance and rel_tolerance:
                _pd.testing.assert_frame_equal(df_new, df_expected, atol=abs_tolerance, rtol=rel_tolerance,
                                               check_dtype=False, check_exact=False)
            elif abs_tolerance:
                _pd.testing.assert_frame_equal(df_new, df_expected, atol=abs_tolerance, check_dtype=False,
                                               check_exact=False)
            elif rel_tolerance:
                _pd.testing.assert_frame_equal(df_new, df_expected, rtol=rel_tolerance, check_dtype=False,
                                               check_exact=False)
            else:
                _pd.testing.assert_frame_equal(df_new, df_expected, check_dtype=False, check_exact=True)
        except AssertionError as current_error_2:
            errors.append(current_error)
            errors.append(current_error_2)
            if manual_test:
                """Plot differences in sheet to simplify comparison."""
                plot_dfs_and_differences(df_new, df_expected, sheet)


def check_assertion(testCase: _ut.TestCase, errors: list, actual, expected):
    try:
        testCase.assertEqual(actual, expected)
    except AssertionError as current_error:
        errors.append(current_error)


def check_dataframe_assertion(
        errors: list,
        actual: _pd.DataFrame,
        expected: _pd.DataFrame,
        absolute_tolerance: float = None,
        check_dtype: bool = True,
):
    try:
        if absolute_tolerance:
            _pd.testing.assert_frame_equal(actual, expected, atol=absolute_tolerance, check_dtype=check_dtype)
        else:
            _pd.testing.assert_frame_equal(actual, expected, atol=absolute_tolerance, check_dtype=check_dtype)
    except AssertionError as current_error:
        errors.append(current_error)


def run_python_script(script_path: _pl.Path) -> None:
    cwd = _os.getcwd()
    _os.chdir(script_path.parent)
    outcome = _sp.run([ROOT_DIR / 'venv' / 'Scripts' / 'python.exe', script_path, '/H'],
                      shell=True, check=False,  # check=True does not provide helpful feedback.
                      capture_output=True, text=True)
    _os.chdir(cwd)

    if outcome.returncode != 0:
        raise AssertionError(outcome.stderr)


def import_example_script(dir_path: _pl.Path, file_name: str):
    """
    TODO: Typing of module???
    """
    _sys.path.append(str(dir_path))
    return _ilib.import_module(file_name)




from oemof.tools import economics
from optihood.converters import *
from optihood.sources import PV
from optihood.storages import *
from optihood.sinks import SinkRCModel
from optihood.links import LinkTemperatureDemand
from optihood._helpers import *
import optihood.entities as _ent

intRate = 0.05

class Building:
    def __init__(self, label):
        self.__nodesList = []
        self.__inputs = []
        self.__technologies = []
        self.__costParam = {}
        self.__envParam = {}
        self.__busDict = {}
        self.__buildingLabel = label
        self.__linkBuses = []
        self.__heatSourceSinkBuses = ["heatSourceBus", "heatSinkBus"]


    def getBuildingLabel(self):
        return self.__buildingLabel

    def getNodesList(self):
        return self.__nodesList

    def getBusDict(self):
        return self.__busDict

    def getInputs(self):
        return self.__inputs

    def getTechnologies(self):
        return self.__technologies

    def getCostParam(self):
        return self.__costParam

    def getEnvParam(self):
        return self.__envParam

    def addBus(self, data, opt, mergeLinkBuses, mergeHeatSourceSink, electricityImpact, clusterSize, includeCarbonBenefits):
        # Create Bus objects from buses table
        for i, b in data.iterrows():
                if (mergeLinkBuses and b["label"] in self.__linkBuses) or (mergeHeatSourceSink and b["label"] in self.__heatSourceSinkBuses):
                    label = b["label"]
                else:
                    label = b["label"] + '__' + self.__buildingLabel

                if ((not mergeLinkBuses and not mergeHeatSourceSink) or (mergeLinkBuses and (self.__buildingLabel=='Building1' or label not in self.__linkBuses))) or ((not mergeHeatSourceSink and not mergeLinkBuses) or (mergeHeatSourceSink and (self.__buildingLabel=='Building1' or label not in self.__heatSourceSinkBuses))):
                    bus = solph.Bus(label=label)
                    self.__nodesList.append(bus)
                    self.__busDict[label] = bus
                    if opt=='costs':
                        if not clusterSize: varcost=float(b["excess costs"])
                        else:
                            varcost = None
                            for d in clusterSize:
                                temp = pd.Series(np.tile([float(b["excess costs"])*clusterSize[d]], 24))
                                if varcost is not None:
                                    varcost = varcost._append(temp, ignore_index=True)
                                else:
                                    varcost = temp
                    elif opt=='env' and includeCarbonBenefits:varcost=-electricityImpact['impact']
                    else: varcost=0
                    if b["excess"]:
                        self.__nodesList.append(
                            solph.components.Sink(
                                label="excess"+label,
                                inputs={
                                    self.__busDict[label]: solph.Flow(
                                        variable_costs=varcost
                                    )}))
                        # add the excess production cost to self.__costParam
                        self.__costParam["excess"+label] = float(b["excess costs"])
                    if "shortage" in b:
                        if b["shortage"]:
                            self.__nodesList.append(
                                solph.components.Source(
                                    label="shortage"+label,
                                    outputs={
                                        self.__busDict[label]: solph.Flow(
                                            variable_costs=float(b["shortage costs"])*(opt == "costs")  # if opt = "env" variable costs should be zero
                                        )}))
                            # add the excess production cost to self.__costParam
                            self.__costParam["shortage"+label] = float(b["shortage costs"])

        if (mergeLinkBuses or mergeHeatSourceSink) and self.__buildingLabel=='Building1':
            return self.__busDict

    def addToBusDict(self, busDictBuilding1, mergeType):
        if mergeType=='links':
            buses = self.__linkBuses
        elif mergeType =='heatSourceSink':
            buses = self.__heatSourceSinkBuses
        for label in buses:
            self.__busDict[label] = busDictBuilding1[label]

    def linkBuses(self, busesToMerge):
        for b in busesToMerge:
            if b == "electricity":
                self.__linkBuses.extend(["electricityBus", "electricityInBus"])
            elif b == "space_heat":
                self.__linkBuses.extend(["spaceHeatingBus", "shDemandBus"])
            elif b == "domestic_hot_water":
                self.__linkBuses.extend(["domesticHotWaterBus", "dhwDemandBus"])
            elif b == "heat_buses":
                self.__linkBuses.extend(["heatDemandBus0", "heatDemandBus2"])
            else:
                self.__linkBuses.append(b)

    def addPV(self, data, data_timeseries, opt, dispatchMode):
        # Create Source objects from table 'commodity sources'
        for i, s in data.iterrows():
                if opt == "costs":
                    epc=self._calculateInvest(s)[0]
                    base=self._calculateInvest(s)[1]
                    env_capa=float(s["impact_cap"]) / float(s["lifetime"])
                    env_flow=float(s["elec_impact"])
                    varc=0 # variable cost is only passed for environmental optimization if there are emissions per kWh of energy produced from the unit

                    envParam = [0, float(s["elec_impact"]), float(s["impact_cap"]) / float(s["lifetime"])]

                elif opt == "env":
                    epc = float(s["impact_cap"]) / float(s["lifetime"])
                    base = 0
                    env_capa = float(s["impact_cap"]) / float(s["lifetime"])
                    env_flow = float(s["elec_impact"])
                    varc = float(s["elec_impact"]) # variable cost is only passed for environmental optimization if there are emissions per kWh of energy produced from the unit

                    envParam = [0, float(s["elec_impact"]), float(s["impact_cap"]) / float(s["lifetime"])]

                # If roof area and zenith do not exist in the excel file
                if 'roof_area' not in s.keys():
                    s["roof_area"] = np.nan
                if 'zenith_angle' not in s.keys():
                    s["zenith_angle"] = np.nan
                if 'efficiency' not in s.keys():
                    s["efficiency"] = np.nan

                self.__nodesList.append(PV(s["label"], self.__buildingLabel,
                                           self.__busDict[s["to"] + '__' + self.__buildingLabel],
                                           float(s["peripheral_losses"]), float(s["latitude"]), float(s["longitude"]),
                                           float(s["tilt"]), float(s["efficiency"]), s["roof_area"],
                                           s["zenith_angle"], s["azimuth"],
                                           data_timeseries['gls'],
                                           data_timeseries['str.diffus'],
                                           data_timeseries['tre200h0'], float(s["capacity_min"]), float(s["capacity_max"]),
                                           epc, base, env_capa, env_flow, varc, dispatchMode))

                self.__envParam[s["label"] + '__' + self.__buildingLabel] = envParam

                self.__costParam[s["label"] + '__' + self.__buildingLabel] = [self._calculateInvest(s)[0],
                                                                              self._calculateInvest(s)[1]]
                self.__technologies.append(
                    [s["to"] + '__' + self.__buildingLabel, s["label"] + '__' + self.__buildingLabel])

    def addSolar(self, data, data_timeseries, opt, mergeLinkBuses, dispatchMode, temperatureLevels):
        # Create Source objects from table 'commodity sources'
        for i, s in data.iterrows():
                if mergeLinkBuses and s["from"] in self.__linkBuses:
                    inputBusLabel = s["from"]
                else:
                    inputBusLabel = s["from"] + '__' + self.__buildingLabel
                outputBuses = [self.__busDict[o + '__' + self.__buildingLabel] for o in s["to"].split(",")]
                connectBuses = [self.__busDict[c + '__' + self.__buildingLabel] for c in s["connect"].split(",")]
                if isinstance(s["delta_temp_n"], float) or isinstance(s["delta_temp_n"], int):
                    delta_temp_n = [float(s["delta_temp_n"])]
                else:
                    delta_temp_n = [float(t) for t in s["delta_temp_n"].split(",")]
                if opt == "costs":
                    epc=self._calculateInvest(s)[0]
                    base=self._calculateInvest(s)[1]
                    env_capa=float(s["impact_cap"]) / float(s["lifetime"])
                    env_flow=float(s["heat_impact"])
                    varc=0 # variable cost is only passed for environmental optimization if there are emissions per kWh of energy produced from the unit

                    envParam = [env_flow, 0, env_capa]

                elif opt == "env":
                    epc=float(s["impact_cap"]) / float(s["lifetime"])
                    base=0
                    env_capa= float(s["impact_cap"]) / float(s["lifetime"])
                    env_flow=float(s["heat_impact"])
                    varc= float(s["heat_impact"]) # variable cost is only passed for environmental optimization if there are emissions per kWh of energy produced from the unit

                    envParam = [env_flow, 0, env_capa]

                # If roof area and zenith do not exist in the excel file
                if 'roof_area' not in s.keys():
                    s["roof_area"] = np.nan
                if 'zenith_angle' not in s.keys():
                    s["zenith_angle"] = np.nan
                collector=SolarCollector(s["label"], self.__buildingLabel,
                                                       self.__busDict[inputBusLabel],
                                                       outputBuses,
                                                       connectBuses,
                                                       float(s["electrical_consumption"]), float(s["peripheral_losses"]), float(s["latitude"]),
                                                       float(s["longitude"]), float(s["tilt"]), s["roof_area"],
                                                       s["zenith_angle"], float(s["azimuth"]),
                                                       float(s["eta_0"]), float(s["a_1"]), float(s["a_2"]), float(s["temp_collector_inlet"]),
                                                       delta_temp_n, data_timeseries['gls'], data_timeseries['str.diffus'],
                                                        data_timeseries['tre200h0'], float(s["capacity_min"]), float(s["capacity_max"]),
                                                       epc, base, env_capa, env_flow, varc, dispatchMode)

                nodes = []
                for t in ["source", "transformer", "sink"]:
                    sh, T2, dhw = collector.getSolar(t)
                    if outputBuses.__len__() == 2:
                        nodes.extend([sh, dhw])
                    elif outputBuses.__len__() == 3:
                        nodes.extend([sh, T2, dhw])
                    else:
                        nodes.extend([sh])
                for x in nodes:
                    self.__nodesList.append(x)

                self.__envParam['heatSource_SH' + s['label'] + '__' + self.__buildingLabel] = envParam
                self.__costParam['heatSource_SH' + s['label'] + '__' + self.__buildingLabel] = [self._calculateInvest(s)[0],
                                                                                                self._calculateInvest(s)[1]]
                self.__technologies.append(
                    [outputBuses[0], s["label"] + 'SH__' + self.__buildingLabel])

                if outputBuses.__len__() >= 2:
                    self.__envParam['heatSource_DHW' + s['label'] + '__' + self.__buildingLabel] = [env_flow, 0, 0]
                    self.__costParam['heatSource_DHW' + s['label'] + '__' + self.__buildingLabel] = [0, 0]
                    self.__technologies.append(
                        [outputBuses[-1], s["label"] + 'DHW__' + self.__buildingLabel])

                if outputBuses.__len__() == 3:
                    self.__envParam['heatSource_T2' + s['label'] + '__' + self.__buildingLabel] = [env_flow, 0, 0]
                    self.__costParam['heatSource_T2' + s['label'] + '__' + self.__buildingLabel] = [0, 0]
                    self.__technologies.append(
                        [outputBuses[1], s["label"] + 'T2__' + self.__buildingLabel])

    def addPVT(self, data, data_timeseries, opt, mergeLinkBuses, dispatchMode):
        # Create Source objects from table 'commodity sources'
        for i, s in data.iterrows():
                if mergeLinkBuses and s["from"] in self.__linkBuses:
                    inputBusLabel = s["from"]
                else:
                    inputBusLabel = s["from"] + '__' + self.__buildingLabel
                outputBuses = [self.__busDict[o + '__' + self.__buildingLabel] for o in s["to"].split(",")]
                connectBuses = [self.__busDict[c + '__' + self.__buildingLabel] for c in s["connect"].split(",")]
                if isinstance(s["delta_temp_n"], float) or isinstance(s["delta_temp_n"], int):
                    delta_temp_n = [float(s["delta_temp_n"])]
                else:
                    delta_temp_n = [float(t) for t in s["delta_temp_n"].split(",")]
                if opt == "costs":
                    epc = self._calculateInvest(s)[0]
                    base = self._calculateInvest(s)[1]
                    env_capa = float(s["impact_cap"]) / float(s["lifetime"])
                    env_flow = float(s["heat_impact"])
                    varc = 0  # variable cost is only passed for environmental optimization if there are emissions per kWh of energy produced from the unit

                    envParam = [env_flow, 0, env_capa]

                elif opt == "env":
                    epc = float(s["impact_cap"]) / float(s["lifetime"])
                    base = 0
                    env_capa = float(s["impact_cap"]) / float(s["lifetime"])
                    env_flow = float(s["heat_impact"])
                    varc = float(s[
                                     "heat_impact"])  # variable cost is only passed for environmental optimization if there are emissions per kWh of energy produced from the unit

                    envParam = [env_flow, 0, env_capa]

                # If roof area does not exist in the excel file
                if 'roof_area' not in s.keys():
                    s["roof_area"] = np.nan
                else:
                    s["roof_area"] = float(s["roof_area"])
                pvtcollector = PVT(s["label"], self.__buildingLabel,
                                                       self.__busDict[inputBusLabel],
                                                       outputBuses,
                                                       connectBuses,
                                                       float(s["electrical_consumption"]), float(s["peripheral_losses"]), float(s["latitude"]),
                                                       float(s["longitude"]), float(s["tilt"]), s["roof_area"],
                                                       float(s["zenith_angle"]), float(s["azimuth"]),
                                                       float(s["eta_0"]), float(s["a_1"]), float(s["a_2"]), float(s["temp_collector_inlet"]),data_timeseries['tre200h0'],
                                                       delta_temp_n, data_timeseries['gls'], data_timeseries['str.diffus'], float(s["capacity_min"]), float(s["capacity_max"]),
                                                       epc, base, env_capa, env_flow, varc, dispatchMode)
                nodes = [pvtcollector.getPVT("el_source")]
                for t in ["heat_source", "heat_transformer", "excess_heat_sink"]:
                    sh, T2, dhw = pvtcollector.getPVT(t)
                    if outputBuses.__len__() == 3:
                        nodes.extend([sh, dhw])
                    elif outputBuses.__len__() == 4:
                        nodes.extend([sh, T2, dhw])
                    else:
                        nodes.extend([sh])
                for x in nodes:
                    self.__nodesList.append(x)

                self.__envParam['heatSource_SH' + s['label'] + '__' + self.__buildingLabel] = envParam
                self.__costParam['heatSource_SH' + s['label'] + '__' + self.__buildingLabel] = [self._calculateInvest(s)[0],
                                                                                                self._calculateInvest(s)[1]]
                self.__technologies.append(
                    [outputBuses[0], s["label"] + 'SH__' + self.__buildingLabel])

                if outputBuses.__len__() >= 3:
                    self.__envParam['heatSource_DHW' + s['label'] + '__' + self.__buildingLabel] = [env_flow, 0, 0]
                    self.__costParam['heatSource_DHW' + s['label'] + '__' + self.__buildingLabel] = [0, 0]
                    self.__technologies.append(
                        [outputBuses[-2], s["label"] + 'DHW__' + self.__buildingLabel])

                if outputBuses.__len__() == 4:
                    self.__envParam['heatSource_T2' + s['label'] + '__' + self.__buildingLabel] = [env_flow, 0, 0]
                    self.__costParam['heatSource_T2' + s['label'] + '__' + self.__buildingLabel] = [0, 0]
                    self.__technologies.append(
                        [outputBuses[1], s["label"] + 'T2__' + self.__buildingLabel])


    def addGridSeparation(self, dataGridSeparation, mergeLinkBuses):
        if not dataGridSeparation.empty:
            for i, gs in dataGridSeparation.iterrows():
                if mergeLinkBuses:
                    if gs["label"] in ['gridElectricity','electricitySource','shSource']:
                        label = gs["label"]+'__'+self.__buildingLabel
                    else:
                        label = gs["label"]
                    if gs["from"] in self.__linkBuses:
                        inputBusLabel = gs["from"]
                    else:
                        inputBusLabel = gs["from"]+'__'+self.__buildingLabel
                    if gs["to"] in self.__linkBuses:
                        outputBusLabel = gs["to"]
                    else:
                        outputBusLabel = gs["to"]+'__'+self.__buildingLabel
                else:
                    label = gs["label"]+'__'+self.__buildingLabel
                    inputBusLabel = gs["from"]+'__'+self.__buildingLabel
                    outputBusLabel = gs["to"] + '__' + self.__buildingLabel

                if (self.__buildingLabel in label) or (mergeLinkBuses and self.__buildingLabel=='Building1'):
                    self.__nodesList.append(solph.components.Transformer(label=label,
                                                              inputs={self.__busDict[inputBusLabel]: solph.Flow()},
                                                              outputs={self.__busDict[outputBusLabel]: solph.Flow()},
                                                      conversion_factors={self.__busDict[outputBusLabel]: float(gs["efficiency"])}))

    def addSource(self, data, data_elimpact, data_elcost, data_natGascost, data_natGasImpact, timeseries, opt, mergeHeatSourceSink, mergeLinkBuses):
        # Create Source objects from table 'commodity sources'

        for i, cs in data.iterrows():
                label = cs["label"]+'__' + self.__buildingLabel
                if (mergeHeatSourceSink and cs["to"] in self.__heatSourceSinkBuses) or \
                        (mergeLinkBuses and cs["to"] in self.__linkBuses):
                    outputBusLabel = cs["to"]
                else:
                    outputBusLabel = cs["to"] + '__' + self.__buildingLabel
                if opt == "costs":
                    if 'electricity' in label:
                        varCosts = data_elcost["cost"]
                    elif 'naturalGas' in label:
                        varCosts = data_natGascost["cost"]
                    else:
                        varCosts = float(cs["variable costs"])
                elif 'electricity' in label:
                    varCosts = data_elimpact["impact"]
                elif 'naturalGas' in label:
                    varCosts = data_natGasImpact["impact"]
                else:
                    varCosts = float(cs["CO2 impact"])

                if 'electricity' in label:
                    envImpactPerFlow = data_elimpact["impact"]
                    envParameter = data_elimpact["impact"]
                    costParameter = data_elcost["cost"]
                elif 'naturalGas' in label:
                    envImpactPerFlow = data_natGasImpact["impact"]
                    envParameter = data_natGasImpact["impact"]
                    costParameter = data_natGascost["cost"]
                else:
                    envImpactPerFlow = float(cs["CO2 impact"])
                    envParameter = float(cs["CO2 impact"])
                    costParameter = float(cs["variable costs"])
                    # add the inputs (natural gas, wood, etc...) to self.__inputs
                    self.__inputs.append([label, outputBusLabel])
                flowargs = {'variable_costs':varCosts,
                            'custom_attributes': {'env_per_flow': envImpactPerFlow},}
                if "fixed" in cs:
                    if pd.notna(cs["fixed"]) and int(cs["fixed"])==1:
                        if "waste" in label.lower():
                            fixed_profile = timeseries[f"waste__{self.__buildingLabel}"]
                        flowargs.update({"fix": fixed_profile["fixed_source"], "nominal_value": float(cs["nominal_value"])})
                elif "potential" in label.lower():
                    flowargs.update({'full_load_time_max': float(cs["full_load_time_max"]), "nominal_value": float(cs["nominal_value"])})

                self.__nodesList.append(solph.components.Source(
                    label=label,
                    outputs={self.__busDict[outputBusLabel]: solph.Flow(**flowargs)}))

                # set environment and cost parameters
                self.__envParam[label] = envParameter
                self.__costParam[label] = costParameter

    def addSink(self, data, timeseries, buildingModelParams, mergeLinkBuses, mergeHeatSourceSink, temperatureLevels):
        # Create Sink objects with fixed time series from 'demand' table
        for i, de in data.iterrows():
                sinkLabel = de["label"]+'__'+self.__buildingLabel
                if (mergeLinkBuses and de["from"] in self.__linkBuses) or (mergeHeatSourceSink and de["from"] in self.__heatSourceSinkBuses):
                    inputBusLabel = de["from"]
                elif temperatureLevels and "," in de["from"]:
                    inputBusLabel = [bus+'__'+self.__buildingLabel for bus in de["from"].split(",")]
                    weights = {self.__busDict[b]: float(w) for b,w in zip(inputBusLabel, de["weight"].split(","))}
                else:
                    inputBusLabel = de["from"]+'__'+self.__buildingLabel

                if de["building model"] == 'Yes':   # Should a building model be used?
                    # Only valid for SH demands at the moment
                    if "spaceHeatingDemand" not in sinkLabel:
                        logging.error("Building model has been selected for one or more demands other than space heating. "
                                      "This is not supported in optihood.")
                    if temperatureLevels and "," in de["from"]:
                        inputBusDict = {self.__busDict[k]: solph.Flow() for k in inputBusLabel}
                    else:
                        inputBusDict = {self.__busDict[inputBusLabel]: solph.Flow()}
                    # create sink
                    # Building model with a possibility of output bus
                    outputBusLabel = de['building model out']+'__'+self.__buildingLabel
                    args = {'label': sinkLabel,
                            'inputs': inputBusDict,
                            'outputs': {self.__busDict[outputBusLabel]: solph.Flow()},
                            'tAmbient': buildingModelParams["timeseries"]['tAmb'].values,
                            'totalIrradiationHorizontal': buildingModelParams["timeseries"]['IrrH'].values,
                            'heatGainOccupants': buildingModelParams["timeseries"]["Qocc"].values,
                            }
                    paramList = ['gAreaWindows', 'rDistribution', 'cDistribution', 'rWall', 'cWall', 'rIndoor',
                              'cIndoor', 'qDistributionMin', 'qDistributionMax', 'tIndoorMin', 'tIndoorMax',
                              'tIndoorInit', 'tWallInit', 'tDistributionInit']
                    for param in paramList:
                        if 'tIndoorSet' in buildingModelParams['timeseries'].columns and param in ['tIndoorMin', 'tIndoorMax']:
                            continue
                        if buildingModelParams[param] != '' and param != 'tIndoorMin':
                            args.update({param:buildingModelParams[param]})
                        elif param == 'tIndoorMin':
                            tIndoorNight = 20
                            tIndoorDay = 22
                            tlow = [tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight,
                                  tIndoorNight, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay,
                                  tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay, tIndoorDay,
                                  tIndoorNight, tIndoorNight, tIndoorNight, tIndoorNight]*365
                            args.update({'tIndoorMin': np.array(tlow)})
                    self.__nodesList.append(SinkRCModel(**args))
                else:
                    # set static inflow values, if any
                    inflow_args = {}
                    if int(de["fixed"])==1:
                        inflow_args["nominal_value"] = float(de["nominal value"])
                        # get time series for node and parameter
                        for col in timeseries.columns.values:
                            if col == de["label"]:
                                inflow_args["fix"] = timeseries[col]
                                if temperatureLevels and "," in de["from"]:
                                    demandInputLabel = col + "Bus" + '__' + self.__buildingLabel
                                    bus = solph.Bus(label=demandInputLabel)
                                    self.__nodesList.append(bus)
                                    self.__busDict[demandInputLabel] = bus
                        if temperatureLevels and "," in de["from"]:
                            inputDummyBusDict = {self.__busDict[k]: solph.Flow() for k in inputBusLabel}
                            self.__nodesList.append(LinkTemperatureDemand(
                                label="dummy_"+sinkLabel,
                                inputs=inputDummyBusDict,
                                outputs={self.__busDict[demandInputLabel]: solph.Flow()},
                                conversion_factors={k: weights[k] for k in inputDummyBusDict}
                            ))
                            inputBusDict = {self.__busDict[demandInputLabel]: solph.Flow(**inflow_args)}
                        else:
                            inputBusDict = {self.__busDict[inputBusLabel]: solph.Flow(**inflow_args)}
                        # create sink
                        self.__nodesList.append(
                        solph.components.Sink(
                                label=sinkLabel,
                                inputs=inputBusDict,
                            )
                        )

    def _addHeatPump(self, data, operationTempertures, temperature_evap, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode):
        self._add_hp_or_chiller(data, operationTempertures, temperature_evap, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode, Component=HeatPumpLinear)

    def _add_chiller(self, data, operationTempertures, temperature_evap, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode):
        self._add_hp_or_chiller(data, operationTempertures, temperature_evap, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode, Component=Chiller)

    def _add_hp_or_chiller(self, data, operationTempertures, temperature_evap, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode,
                           Component: HeatPumpLinear | Chiller = HeatPumpLinear,
                           ):
        """
        HPs and chillers are implemented the same way.
        The required class is passed in as a 'Component'
        """
        label = LabelStringManipulator(data[_ent.TransformerLabels.label] + '__' + self.__buildingLabel)
        from_bus = _ent.TransformerLabels.from_bus
        to_bus = _ent.TransformerLabels.to
        if mergeLinkBuses and data[from_bus] in self.__linkBuses:
            inputBusLabel = [data[from_bus]]
        elif mergeLinkBuses and any([b in self.__linkBuses for b in data[from_bus].split(',')]):
            inputBusLabel = [i + '__' + self.__buildingLabel for i in data[from_bus].split(",") if
                             i not in self.__linkBuses]
            inputBusLabel.extend([i for i in data[from_bus].split(",") if i in self.__linkBuses])
        elif mergeHeatSourceSink and any([b in self.__heatSourceSinkBuses for b in data[from_bus].split(',')]):
            inputBusLabel = [i + '__' + self.__buildingLabel for i in data[from_bus].split(",") if
                             i not in self.__heatSourceSinkBuses]
            inputBusLabel.extend([i for i in data[from_bus].split(",") if i in self.__heatSourceSinkBuses])
        else:
            inputBusLabel = [i + '__' + self.__buildingLabel for i in data[from_bus].split(",")]
        inputBuses = [self.__busDict[i] for i in inputBusLabel]
        outputBusLabels = [i + '__' + self.__buildingLabel for i in data[to_bus].split(",")]
        outputBuses = [self.__busDict[i] for i in outputBusLabels]
        envImpactPerCapacity = float(data[_ent.TransformerLabels.impact_cap]) / float(data[_ent.TransformerLabels.lifetime])
        if data[_ent.TransformerLabels.capacity_min] == 'x':
            capacityMinSH = float(data[_ent.TransformerLabels.capacity_SH])
        else:
            capacityMinSH = float(data[_ent.TransformerLabels.capacity_min])
        if pattern_at_start_followed_by_number(_ent.TransformerTypes.HP, label.prefix):
            # default coefficients for air-source heat pump
            coef_W = [0.66610, -2.2365, 15.541, 25.705, -17.407, 3.8145]
            coef_Q = [11.833, 96.504, 14.496, -50.064, 161.02, -133.60]
        elif pattern_at_start_followed_by_number(_ent.TransformerTypes.GWHP, label.prefix):
            # default coefficients for geothermal heat pump
            coef_W = [0.1600, -1.2369, 19.9391, 19.3448, 7.1057, -1.4048]
            coef_Q = [13.8978, 114.8358, -9.3634, -179.4227, 342.3363, -12.4969]
        elif pattern_at_start_followed_by_number(_ent.TransformerTypes.Chiller, label.prefix):
            # default coefficients for chiller
            coef_W = [6.534, 6.4908, 0.3555, 2.578, 8.076, -11.005]
            coef_Q = [79.897, 738.27, -22.540, -1499.7, 2266.1, -462.81]
        if (_ent.HeatPumpCoefficientLabels.coef_W in data) and (_ent.HeatPumpCoefficientLabels.coef_Q in data):
            #comma-separated entries for user-defined coefficients split into a list
            coef_W = [float(c) for c in data[_ent.HeatPumpCoefficientLabels.coef_W].split(",")]
            coef_Q = [float(c) for c in data[_ent.HeatPumpCoefficientLabels.coef_Q].split(",")]
        heatPump = Component(label.full_name, operationTempertures, temperature_evap,
                                  coef_W,
                                  coef_Q,
                                  inputBuses,
                                  outputBuses,
                                  capacityMinSH, float(data[_ent.TransformerLabels.capacity_SH]),float(data[_ent.TransformerLabels.efficiency]),
                                  self._calculateInvest(data)[0] * (opt == "costs") + envImpactPerCapacity*(opt == "env"),
                                  self._calculateInvest(data)[1] * (opt == "costs"),
                                  float(data[_ent.TransformerLabels.heat_impact]) * (opt == "env"),
                                  float(data[_ent.TransformerLabels.heat_impact]), envImpactPerCapacity, dispatchMode)

        self.__nodesList.append(heatPump.getHP("sh"))

        # set technologies, environment and cost parameters
        self.__technologies.extend([[o, label.full_name] for o in outputBusLabels])
        self.__costParam[label.full_name] = [self._calculateInvest(data)[0], self._calculateInvest(data)[1]]

        self.__envParam[label.full_name] = [float(data[_ent.TransformerLabels.heat_impact]), 0, envImpactPerCapacity]

    def _addGeothemalHeatPumpSplit(self, data, operationTemperatures, temperatureAmb, opt, mergeLinkBuses, dispatchMode):
        if len(operationTemperatures)==3:
            logging.error("The transformer type GSHP split is not supported with discrete temperature levels. "
                          "Set temperatureLevels to False or use the transformer GSHP")
        gwhpDHWLabel = data["label"][0:4] + str(operationTemperatures[1]) + '__' + self.__buildingLabel
        gwhpSHLabel = data["label"][0:4] + str(operationTemperatures[0]) + '__' + self.__buildingLabel
        if mergeLinkBuses and data["from"] in self.__linkBuses:
            inputBusLabel = data["from"]
        else:
            inputBusLabel = data["from"] + '__' + self.__buildingLabel
        outputSHBusLabel = data["to"].split(",")[0] + '__' + self.__buildingLabel
        outputDHWBusLabel = data["to"].split(",")[1] + '__' + self.__buildingLabel
        envImpactPerCapacity = float(data["impact_cap"]) / float(data["lifetime"])
        capacityDHW = float(data["capacity_DHW"])
        capacitySH = float(data["capacity_SH"])
        if data["capacity_min"] == 'x':
            capacityMinSH = capacitySH
            capacityMinDHW = capacityDHW
        else:
            capacityMinSH = capacityMinDHW = float(data["capacity_min"])

        geothermalheatPumpSH = GeothermalHeatPumpLinearSingleUse(self.__buildingLabel, operationTemperatures[0], temperatureAmb,
                                  self.__busDict[inputBusLabel],
                                  self.__busDict[outputSHBusLabel],
                                  capacityMinSH, capacitySH,
                                  self._calculateInvest(data)[0] * (opt == "costs") + envImpactPerCapacity*(opt == "env"),
                                  self._calculateInvest(data)[1] * (opt == "costs"),
                                  float(data["heat_impact"]) * (opt == "env"),
                                  float(data["heat_impact"]), envImpactPerCapacity, dispatchMode)
        geothermalheatPumpDHW = GeothermalHeatPumpLinearSingleUse(self.__buildingLabel, operationTemperatures[1],
                                                        temperatureAmb,
                                                        self.__busDict[inputBusLabel],
                                                        self.__busDict[outputDHWBusLabel],
                                                        capacityMinDHW, capacityDHW,
                                                        self._calculateInvest(data)[0] * (
                                                                    opt == "costs") + envImpactPerCapacity * (
                                                                    opt == "env"),
                                                        self._calculateInvest(data)[1] * (opt == "costs"),
                                                        float(data["heat_impact"]) * (opt == "env"),
                                                        float(data["heat_impact"]), envImpactPerCapacity, dispatchMode)

        self.__nodesList.append(geothermalheatPumpSH.getHP("sh"))
        self.__nodesList.append(geothermalheatPumpDHW.getHP("sh"))

        # set technologies, environment and cost parameters
        self.__technologies.append([outputDHWBusLabel, gwhpDHWLabel])
        self.__technologies.append([outputSHBusLabel, gwhpSHLabel])

        self.__costParam[gwhpSHLabel] = [self._calculateInvest(data)[0], self._calculateInvest(data)[1]]
        self.__costParam[gwhpDHWLabel] = [self._calculateInvest(data)[0], self._calculateInvest(data)[1]]

        self.__envParam[gwhpSHLabel] = [float(data["heat_impact"]), 0, envImpactPerCapacity]
        self.__envParam[gwhpDHWLabel] = [float(data["heat_impact"]), 0, envImpactPerCapacity]

    def _addCHP(self, data, timesteps, opt, dispatchMode):
        chpLabel = data["label"] + '__' + self.__buildingLabel
        inputBusLabel = data["from"] + '__' + self.__buildingLabel
        outputBuses = [self.__busDict[o + '__' + self.__buildingLabel] for o in data["to"].split(",")]
        efficiency = [float(e) for e in data["efficiency"].split(",")]
        envImpactPerCapacity = float(data["impact_cap"]) / float(data["lifetime"])
        if data["capacity_min"] == 'x':
            capacityMinSH = float(data["capacity_SH"])
        else:
            capacityMinSH = float(data["capacity_min"])
        chp = CHP(self.__buildingLabel, self.__busDict[inputBusLabel],
                  outputBuses,
                  efficiency,
                  capacityMinSH,
                  float(data["capacity_SH"]),
                  self._calculateInvest(data)[0] * (opt == "costs") + envImpactPerCapacity * (opt == "env"),
                  self._calculateInvest(data)[1] * (opt == "costs"), float(data["elec_impact"]) * (opt == "env"),
                  float(data["heat_impact"]) * (opt == "env"),
                  float(data["elec_impact"]), float(data["heat_impact"]), envImpactPerCapacity, timesteps, dispatchMode)

        self.__nodesList.append(chp.getCHP())

        # set technologies, environment and cost parameters
        for i in range(len(outputBuses)):
            self.__technologies.append([data["to"].split(",")[i] + '__' + self.__buildingLabel, chpLabel])

        self.__costParam[chpLabel] = [self._calculateInvest(data)[0], self._calculateInvest(data)[1]]

        self.__envParam[chpLabel] = [float(data["heat_impact"]), float(data["elec_impact"]), envImpactPerCapacity]

    def _addBoiler(self, data, opt, dispatchMode):
        label = data["label"] + '__' + self.__buildingLabel
        inputBusLabel = data["from"] + '__' + self.__buildingLabel
        outputBuses = [self.__busDict[o + '__' + self.__buildingLabel] for o in data["to"].split(",")]
        efficiency = [float(e) for e in str(data["efficiency"]).split(",")]
        envImpactPerCapacity = float(data["impact_cap"]) / float(data["lifetime"])
        if data["capacity_min"] == 'x':
            capacityMinSH = float(data["capacity_SH"])
        else:
            capacityMinSH = float(data["capacity_min"])

        self.__nodesList.append(GenericCombinedTransformer(label, self.__busDict[inputBusLabel],
                  outputBuses,
                  efficiency, capacityMinSH, float(data["capacity_SH"]),
                  self._calculateInvest(data)[0] * (opt == "costs") + envImpactPerCapacity * (opt == "env"),
                  self._calculateInvest(data)[1] * (opt == "costs"), float(data["heat_impact"]) * (opt == "env"), float(data["heat_impact"]), envImpactPerCapacity, dispatchMode))

        # set technologies, environment and cost parameters
        for i in range(len(outputBuses)):
            self.__technologies.append([data["to"].split(",")[i] + '__' + self.__buildingLabel, label])

        self.__costParam[label] = [self._calculateInvest(data)[0], self._calculateInvest(data)[1]]

        self.__envParam[label] = [float(data["heat_impact"]), 0, envImpactPerCapacity]

    def _addElectricRod(self, data, opt, mergeLinkBuses, dispatchMode, temperatureLevels):
        elRodLabel = data["label"] + '__' + self.__buildingLabel
        if mergeLinkBuses and data["from"] in self.__linkBuses:
            inputBusLabel = data["from"]
        else:
            inputBusLabel = data["from"] + '__' + self.__buildingLabel
        outputBuses = [self.__busDict[o + '__' + self.__buildingLabel] for o in data["to"].split(",")]
        efficiency = [float(e) for e in data["efficiency"].split(",")]
        envImpactPerCapacity = float(data["impact_cap"]) / float(data["lifetime"])
        if data["capacity_min"] == 'x':
            capacityMinSH = float(data["capacity_SH"])
        else:
            capacityMinSH = float(data["capacity_min"])

        self.__nodesList.append(GenericCombinedTransformer(elRodLabel, self.__busDict[inputBusLabel],
                                          outputBuses,
                                          efficiency, capacityMinSH, float(data["capacity_SH"]),
                                          self._calculateInvest(data)[0] * (opt == "costs") + envImpactPerCapacity * (
                                                      opt == "env"),
                                          self._calculateInvest(data)[1] * (opt == "costs"),
                                          float(data["heat_impact"]) * (opt == "env"), float(data["heat_impact"]),
                                          envImpactPerCapacity, dispatchMode))

        # set technologies, environment and cost parameters
        for i in range(len(outputBuses)):
            self.__technologies.append([data["to"].split(",")[i] + '__' + self.__buildingLabel, elRodLabel])

        self.__costParam[elRodLabel] = [self._calculateInvest(data)[0], self._calculateInvest(data)[1]]
        self.__envParam[elRodLabel] = [float(data["heat_impact"]), 0, envImpactPerCapacity]

    def _addGenericTransformer(self, data):
        inputBusLabel = data['from']+ '__Building1'
        outputBusLabels = [b + '__' + self.__buildingLabel for b in data['to'].split(",")]
        outputDict = {self.__busDict[o]:solph.Flow(variable_costs=0, nominal_value=100000000000, custom_attributes={'env_per_flow': 0}) for o in outputBusLabels}
        convFactors = {self.__busDict[o]:1 for o in outputBusLabels}
        self.__nodesList.append(solph.components.Transformer(label=data['label'] + '__Building1',
                                                  inputs={self.__busDict[inputBusLabel]: solph.Flow()},
                                                  outputs=outputDict,
                                                  conversion_factors=convFactors))
    
    def addTransformer(self, data, operationTemperatures, temperatureAmb, temperatureGround, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode, temperatureLevels):
        for i, t in data.iterrows():
            if t["active"]:
                if pattern_at_start_followed_by_number("HP", t["label"]):
                    self._addHeatPump(t, operationTemperatures[0], temperatureAmb, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode)
                elif pattern_at_start_followed_by_number("GWHP", t["label"]):
                    self._addHeatPump(t, operationTemperatures[0], temperatureGround, opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode)
                elif t["label"] == "GWHP split":
                    self._addGeothemalHeatPumpSplit(t, operationTemperatures[0], temperatureGround, opt, mergeLinkBuses, dispatchMode)
                elif "CHP" in t["label"]:
                    self._addCHP(t, len(temperatureAmb), opt, dispatchMode)
                elif "Boiler" in t["label"]:
                    self._addBoiler(t, opt, dispatchMode)
                elif pattern_at_start_followed_by_number("ElectricRod", t["label"]):
                    self._addElectricRod(t, opt, mergeLinkBuses, dispatchMode, temperatureLevels)
                elif pattern_at_start_followed_by_number("Chiller", t["label"]):
                    self._add_chiller(t, [operationTemperatures[1][1]], np.full(len(temperatureGround), operationTemperatures[1][0]), opt, mergeLinkBuses, mergeHeatSourceSink, dispatchMode)
                else:
                    logging.warning("Transformer label not identified, adding generic transformer component...")
                    self._addGenericTransformer(t)

    def addStorage(self, data, storageParams, ambientTemperature, opt, mergeLinkBuses, dispatchMode, temperatureLevels):
        sList = []
        for i, s in data.iterrows():
                label = LabelStringManipulator(s["label"]+'__'+self.__buildingLabel)
                storageLabel = label.full_name
                label_prefix_without_digits = label.strip_trailing_digits_from_prefix()
                if temperatureLevels and "thermalStorage" in storageLabel:
                    inputBuses = [self.__busDict[iLabel + '__' + self.__buildingLabel] for iLabel in s["from"].split(",")]
                    outputBuses = [self.__busDict[oLabel + '__' + self.__buildingLabel] for oLabel in s["to"].split(",")]
                    storTemperatures = storageParams["stratified_storage"].at[label_prefix_without_digits,"temp_h"].split(",")
                    self.__technologies.append([outputBuses[0].label, f'dummy_{storageLabel.replace("thermalStorage",f"thermalStorage{int(storTemperatures[0])}")}'])
                    self.__technologies.append([outputBuses[1].label, storageLabel.replace("thermalStorage",f"thermalStorage{int(storTemperatures[-1])}")])
                else:
                    if mergeLinkBuses and s["from"] in self.__linkBuses:
                        inputBusLabel = s["from"]
                    else:
                        inputBusLabel = s["from"]+'__'+self.__buildingLabel
                    if mergeLinkBuses and s["to"] in self.__linkBuses:
                        outputBusLabel = s["to"]
                    else:
                        outputBusLabel = s["to"]+'__'+self.__buildingLabel
                    self.__technologies.append([outputBusLabel, storageLabel])
                envImpactPerCapacity = float(s["impact_cap"]) / float(s["lifetime"])         # annualized value
                # set technologies, environment and cost parameters
                self.__costParam[storageLabel] = [self._calculateInvest(s)[0], self._calculateInvest(s)[1]]
                self.__envParam[storageLabel] = [float(s["heat_impact"]), float(s["elec_impact"]), envImpactPerCapacity]

                if "electricalStorage" in storageLabel:
                    self.__nodesList.append(ElectricalStorage(self.__buildingLabel, self.__busDict[inputBusLabel],
                                                              self.__busDict[outputBusLabel], float(s["capacity loss"]),
                                                            float(s["initial capacity"]), float(s["efficiency inflow"]),
                                                            float(s["efficiency outflow"]), float(s["capacity min"]),
                                                            float(s["capacity max"]),
                                                            self._calculateInvest(s)[0]*(opt == "costs") + envImpactPerCapacity*(opt == "env"),
                                                            self._calculateInvest(s)[1]*(opt == "costs"),
                                                            float(s["elec_impact"])*(opt == "env"),
                                                            float(s["elec_impact"]), envImpactPerCapacity, dispatchMode))

                if storageLabel.startswith(("dhwStorage", "shStorage", "coolingBufferStorage")) and not temperatureLevels:
                    self.__nodesList.append(ThermalStorage(storageLabel,
                                                           storageParams["stratified_storage"], self.__busDict[inputBusLabel],
                                                           self.__busDict[outputBusLabel],
                                                        float(s["initial capacity"]), float(s["capacity min"]),
                                                        float(s["capacity max"]),
                                                        self._calculateInvest(s)[0]*(opt == "costs") + envImpactPerCapacity*(opt == "env"),
                                                        self._calculateInvest(s)[1]*(opt == "costs"), float(s["heat_impact"])*(opt == "env"),
                                                        float(s["heat_impact"]), envImpactPerCapacity, dispatchMode))
                elif "pitStorage" in storageLabel and not temperatureLevels:
                    self.__nodesList.append(ThermalStoragePit(storageLabel,
                                                           storageParams["stratified_storage"], self.__busDict[inputBusLabel],
                                                           self.__busDict[outputBusLabel],
                                                        float(s["initial capacity"]), float(s["capacity min"]),
                                                        float(s["capacity max"]),
                                                        self._calculateInvest(s)[0]*(opt == "costs") + envImpactPerCapacity*(opt == "env"),
                                                        self._calculateInvest(s)[1]*(opt == "costs"), float(s["heat_impact"])*(opt == "env"),
                                                        float(s["heat_impact"]), envImpactPerCapacity, dispatchMode))

                elif "thermalStorage" in storageLabel and temperatureLevels:
                    storage = ThermalStorageTemperatureLevels(storageLabel,
                               storageParams["stratified_storage"], inputBuses,
                               outputBuses,
                               float(s["initial capacity"]), float(s["capacity min"]),
                               float(s["capacity max"]),
                               self._calculateInvest(s)[0] * (
                                           opt == "costs") + envImpactPerCapacity * (
                                           opt == "env"),
                               self._calculateInvest(s)[1] * (opt == "costs"),
                               float(s["heat_impact"]) * (opt == "env"),
                               float(s["heat_impact"]), envImpactPerCapacity, dispatchMode)
                    sList.append(storage)
                    for i in range(len(inputBuses)):
                        self.__nodesList.append(storage.getStorageLevel(i))
                        self.__nodesList.extend(storage.getDummyComponents(i))
                elif "iceStorage" in storageLabel:
                    if not np.isnan(s["initial capacity"]):
                        initial_ice_frac = float(s["initial capacity"])
                    else:
                        initial_ice_frac = 0

                    self.__nodesList.append(
                        IceStorage(label=storageLabel,
                                   input=self.__busDict[inputBusLabel],
                                   output=self.__busDict[outputBusLabel],
                                   tStorInit=float(s[_ent.StorageLabels.initial_temp]),
                                   fIceInit=initial_ice_frac,
                                   fMax=float(storageParams["ice_storage"].at[label_prefix_without_digits, "max_ice_fraction"]),
                                   rho=float(storageParams["ice_storage"].at[label_prefix_without_digits, "rho_fluid"]),
                                   V=float(s["capacity max"])/1000,  # conversion from L to m3
                                   hfluid=float(storageParams["ice_storage"].at[label_prefix_without_digits, "h_fluid"]),
                                   cp=float(storageParams["ice_storage"].at[label_prefix_without_digits, "cp_fluid"]),
                                   Tamb=ambientTemperature,
                                   UAtank=float(storageParams["ice_storage"].at[label_prefix_without_digits, "UA_tank"]),
                                   inflow_conversion_factor=float(storageParams["ice_storage"].at[label_prefix_without_digits, "inflow_conversion_factor"]),
                                   outflow_conversion_factor=float(storageParams["ice_storage"].at[label_prefix_without_digits, "outflow_conversion_factor"])
                                   )
                    )

                elif (any(storage in storageLabel for storage in ["tankGenericStorage", "boreholeGenericStorage",
                                                               "pitGenericStorage", "aquiferGenericStorage"]) and
                      not temperatureLevels):
                    # These are generic storage models with constant losses per timestep
                    # These models do not include a detailed geometry of the storage type
                    is_tank = False
                    self.__nodesList.append(ThermalStorage(storageLabel,
                           storageParams["stratified_storage"], self.__busDict[inputBusLabel],
                           self.__busDict[outputBusLabel],
                           float(s["initial capacity"]), float(s["capacity min"]),
                           float(s["capacity max"]),
                           self._calculateInvest(s)[0] * (
                                       opt == "costs") + envImpactPerCapacity * (
                                       opt == "env"),
                           self._calculateInvest(s)[1] * (opt == "costs"),
                           float(s["heat_impact"]) * (opt == "env"),
                           float(s["heat_impact"]), envImpactPerCapacity, dispatchMode, is_tank))
                else:
                    logging.error("One of the following issues were encountered: (i) Storage label not identified. Storage label"
                                  "should contain one of the following strings: [electricalStorage, dhwStorage, shStorage, thermalStorage, "
                                  "tankStorage, boreholeStorage, pitStorage]"
                                  "(ii) Separate dhwStorage and/or shStorage selected when temperatureLevels is set as True."
                                  "Either set temperatureLevels to False or rename the storage label to thermalStorage.")
        return sList

    def _calculateInvest(self, data):
        # Calculate the CAPEX and the part of the OPEX not related to energy flows (maintenance)
        c = data["installation"] + data["planification"] + 1
        m = data["maintenance"]
        perCapacity = m * data["invest_cap"] + economics.annuity(c * data["invest_cap"], data["lifetime"], intRate)
        base = m * data["invest_base"] + economics.annuity(c * data["invest_base"], data["lifetime"], intRate)
        return perCapacity, base


from optihood._helpers import *
import oemof.solph as solph
import numpy as np
import pandas as pd
import pvlib


class PV(solph.components.Source):
    def __init__(self, label, buildingLabel, outputs, peripheral_losses, latitude, longitude,
                 pv_tilt, pv_efficiency, roof_area, zenith_angle, pv_azimuth, irradiance_global, irradiance_diffuse, temp_amb_pv, capacityMin,
                 capacityMax, epc, base, env_capa, env_flow, varc, dispatchMode):
        # Creation of a df with 3 columns
        data = self.computePvSolarPosition(irradiance_diffuse, irradiance_global, latitude, longitude, pv_azimuth,
                                           pv_tilt, temp_amb_pv)

        self.pv_electricity = np.minimum(self.pv_precalc(temp_amb_pv, data['pv_ira']/1000), capacityMax)

        if not (np.isnan(roof_area) or np.isnan(zenith_angle) or np.isnan(pv_efficiency)):
            self.surface_used = self._calculateArea(zenith_angle, pv_tilt, pv_azimuth, pv_efficiency)
        else:
            self.surface_used = np.nan
        if dispatchMode:
            investArgs = {'ep_costs':epc,
                         'minimum':capacityMin,
                         'maximum':capacityMax,
                         'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                              'roof_area': roof_area}}
        else:
            investArgs={'ep_costs':epc,
                         'minimum':capacityMin,
                         'maximum':capacityMax,
                         'nonconvex':True,
                         'offset':base,
                         'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                              'roof_area': roof_area}}
        super(PV, self).__init__(label=label + '__' + buildingLabel,
                                 outputs={outputs: solph.Flow(
                                     investment=solph.Investment(**investArgs),
                                     variable_costs=varc,
                                     custom_attributes={'env_per_flow':env_flow},
                                     max=self.pv_electricity
                                 )}
                                 )

    def computePvSolarPosition(self, irradiance_diffuse, irradiance_global, latitude, longitude, pv_azimuth, pv_tilt,
                               temp_amb_pv):
        data = pd.DataFrame(
            {
                'ghi': irradiance_global,
                'dhi': irradiance_diffuse,
                'temp_amb': temp_amb_pv
            }
        )
        solposition = pvlib.solarposition.get_solarposition(
            time=data.index, latitude=latitude, longitude=longitude
        )
        dni = pvlib.irradiance.dni(
            ghi=data['ghi'], dhi=data['dhi'], zenith=solposition['apparent_zenith']
        )
        total_irradiation = pvlib.irradiance.get_total_irradiance(
            surface_tilt=pv_tilt,
            surface_azimuth=pv_azimuth,
            solar_zenith=solposition['apparent_zenith'],
            solar_azimuth=solposition['azimuth'],
            dni=dni.fillna(0),  # fill NaN values with '0'
            ghi=data['ghi'],
            dhi=data['dhi'],
        )
        data['pv_ira'] = total_irradiation['poa_global']
        return data

    # model according to TriHP Energy management algorithms description. D6.2 v1.0
    # nominal power is 1 KW due to the normed optimizer
    def pv_precalc(self, temp_amb, i_H_t, a1=17.23292, a2=0.451708, a3=22.706, a4=-0.062059,
                   a5=0.04277774, a6=9.692792, a7=-1.885868, a8=6.6):

        temp_cell = a1 + a2 * temp_amb + a3 * i_H_t
        pvPower = np.maximum(0, (a4*i_H_t + a5)*temp_cell + a6*i_H_t + a7) / a8
        return pvPower

    def getPV(self):
        return self.__pv

    def _calculateArea(self, zenith_angle, pv_tilt, pv_azimuth, pv_efficiency):
        # Coefficient representing the area used by one unit of capacity of a solar panel
        coeff = -np.sin((zenith_angle+pv_tilt)*np.pi/180)*np.cos(pv_azimuth*np.pi/180)/np.sin(zenith_angle*np.pi/180)/pv_efficiency
        return coeff



from setuptools import setup, find_packages
import subprocess
import pathlib as _pl

with open("README.md", "r") as fh:
    long_description = fh.read()


# Fetch latest tag
# latest_tag = subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"]).strip().decode("utf-8")


def _getInstallRequirements():
    requirementsFile = (
            _pl.Path(__file__).parent / "requirements" / "requirements.in"
    )
    lines = requirementsFile.read_text().split("\n")
    requirements = [l for l in lines if l.strip() and not l.startswith("#")]
    print("")
    print(requirements)
    print("")
    return requirements


setup(
    name='optihood',
    packages=find_packages(),
    version='v0.2.2',
    author="Institute for Solar Technology (SPF), OST Rapperswil",
    author_email="neha.dimri@ost.ch",
    description="optihood optimization framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://optihood.readthedocs.io",
    install_requires=_getInstallRequirements(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.12",
)



import os as _os
import pathlib as _pl
import unittest as _ut

import pandas as _pd
import pytest as _pt

import optihood.MPC.interface as mpci
import optihood.energy_network as en
import optihood.entities as ent
import tests.xls_helpers as xlsh


class TestPrepMpcInputs(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_maybe_get_entries_or_defaults_sh_storage_nothing(self):
        """Ensures nothing is returned, when nothing found."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(0, index=[0, 1], columns=[ent.StorageLabels.label.value])}
        result = mpci.StoragesMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {})

    def test_maybe_get_entries_or_defaults_sh_storage_no_init(self):
        """Checks whether the default is used correctly."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.StorageTypes.shStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["shStorage__B001", 0, 0],
             }
        )}
        result = mpci.StoragesMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {"shStorage__B001": {ent.StorageLabels.initial_capacity.value: 0.0}})

    def test_maybe_get_entries_or_defaults_sh_storage_with_init(self):
        """Checks whether the initial capacity is used correctly."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.StorageTypes.shStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["shStorage__B001", 0, 0],
             ent.StorageLabels.initial_capacity.value: [0.5, 0, 0],
             }
        )}
        result = mpci.StoragesMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {"shStorage__B001": {ent.StorageLabels.initial_capacity.value: 0.5}})

    def test_maybe_get_entries_or_defaults_all_similar_storages_with_init(self):
        """Check whether all cases work together."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.StorageTypes.shStorage, ent.StorageTypes.dhwStorage,
                                            ent.StorageTypes.electricalStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["shStorage__B001", "dhwStorage__B003", "electricalStorage", 0, 0],
             ent.StorageLabels.initial_capacity.value: [0.5, 0.3, 0.75, 0, 0],
             }
        )}
        result = mpci.StoragesMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {'dhwStorage__B003': {'initial capacity': 0.3},
                                      'electricalStorage': {'initial capacity': 0.75},
                                      'shStorage__B001': {'initial capacity': 0.5}}
                             )

    def test_maybe_get_entries_or_defaults_ice_storage_nothing(self):
        """Checks whether this is ignored correctly."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(0, index=[0, 1], columns=[ent.StorageLabels.label.value])}
        result = mpci.IceStorageMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {})

    def test_maybe_get_entries_or_defaults_ice_storage_no_init(self):
        """Checks whether the default is used correctly."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.IceStorageTypes.iceStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["iceStorage__B001", 0, 0],
             }
        )}
        result = mpci.IceStorageMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {"iceStorage__B001": {ent.StorageLabels.initial_capacity.value: 0.0,
                                                           ent.StorageLabels.initial_temp.value: 2.0}})

    def test_maybe_get_entries_or_defaults_ice_storage_with_init(self):
        """Checks whether the initial values are used correctly."""
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.IceStorageTypes.iceStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["iceStorage__B001", 0, 0],
             ent.StorageLabels.initial_capacity.value: [0.5, 0, 0],
             ent.StorageLabels.initial_temp.value: [0.5, 0, 0],
             }
        )}
        result = mpci.IceStorageMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {"iceStorage__B001": {ent.StorageLabels.initial_capacity.value: 0.5,
                                                           ent.StorageLabels.initial_temp.value: 0.5}})

    def test_maybe_get_entries_or_defaults_building_nothing(self):
        """Users should get proper feedback when no data is in the csv."""
        building_model_params = _pd.DataFrame(columns=[ent.BuildingModelParameters.tWallInit])
        nodal_data = {ent.NodeKeysOptional.building_model_parameters: building_model_params}
        with _pt.raises(ValueError):
            mpci.BuildingMPC().maybe_get_entries_or_defaults(nodal_data)

    def test_maybe_get_entries_or_defaults_building_no_init(self):
        """Checks whether the default is used correctly."""
        building_model_params = _pd.DataFrame(
            {ent.BuildingModelParameters.building_unique.value: ["Building_model__B001_0", "Building_model__B001_1",
                                                                 "Building_model__B002_0"],
             }
        )
        nodal_data = {ent.NodeKeysOptional.building_model_parameters: building_model_params}
        result = mpci.BuildingMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {
            'Building_model__B001_0': {'tDistributionInit': 15.0, 'tIndoorInit': 20.0, 'tWallInit': 25.0},
            'Building_model__B001_1': {'tDistributionInit': 15.0, 'tIndoorInit': 20.0, 'tWallInit': 25.0},
            'Building_model__B002_0': {'tDistributionInit': 15.0, 'tIndoorInit': 20.0, 'tWallInit': 25.0}})

    def test_maybe_get_entries_or_defaults_building_with_init(self):
        """Checks whether the default is used correctly."""
        building_model_params = _pd.DataFrame(
            {ent.BuildingModelParameters.building_unique.value: ["Building_model__B001_0", "Building_model__B001_1",
                                                                 "Building_model__B002_0"],
             ent.BuildingModelParameters.tWallInit: [10., 12., 13.],
             ent.BuildingModelParameters.tIndoorInit: [22., 21., 20.3],
             ent.BuildingModelParameters.tDistributionInit: [30., 26., 29.],
             }
        )
        nodal_data = {ent.NodeKeysOptional.building_model_parameters: building_model_params}
        result = mpci.BuildingMPC().maybe_get_entries_or_defaults(nodal_data)
        self.assertDictEqual(result, {
            'Building_model__B001_0': {'tDistributionInit': 30.0, 'tIndoorInit': 22.0, 'tWallInit': 10.0},
            'Building_model__B001_1': {'tDistributionInit': 26.0, 'tIndoorInit': 21.0, 'tWallInit': 12.0},
            'Building_model__B002_0': {'tDistributionInit': 29.0, 'tIndoorInit': 20.3, 'tWallInit': 13.0}})

    def test_prep_mpc_inputs(self):
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.IceStorageTypes.iceStorage, ent.StorageTypes.shStorage,
                                            ent.StorageTypes.dhwStorage,
                                            ent.StorageTypes.electricalStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["iceStorage__B001", "shStorage__B001", "dhwStorage__B003",
                                                   "electricalStorage", 0, 0],
             ent.StorageLabels.initial_capacity.value: [0.25, 0.5, 0.3, 0.75, 0, 0],
             ent.StorageLabels.initial_temp.value: [1.5, "x", "x", "x", "x", "x"],
             }
        )}
        result, label_to_sheet = mpci.prep_mpc_inputs(nodal_data)

        errors = []
        try:
            self.assertDictEqual(result, {'dhwStorage__B003': {'initial capacity': 0.3},
                                          'electricalStorage': {'initial capacity': 0.75},
                                          'shStorage__B001': {'initial capacity': 0.5},
                                          'iceStorage__B001': {'initial capacity': 0.25, 'initial_temp': 1.5},
                                          }
                                 )
        except AssertionError as e:
            errors.append(e)

        try:
            self.assertDictEqual(label_to_sheet,
                                 {'dhwStorage__B003': ent.NodeKeys.storages,
                                  'electricalStorage': ent.NodeKeys.storages,
                                  'iceStorage__B001': ent.NodeKeys.storages,
                                  'shStorage__B001': ent.NodeKeys.storages
                                  })
        except AssertionError as e:
            errors.append(e)

        if errors:
            raise ExceptionGroup(f"Found {len(errors)} issues", errors)

    def test_prep_mpc_inputs_with_building(self):
        nodal_data = {ent.NodeKeys.storages: _pd.DataFrame(
            {ent.CommonLabels.label.value: [ent.IceStorageTypes.iceStorage, ent.StorageTypes.shStorage,
                                            ent.StorageTypes.dhwStorage,
                                            ent.StorageTypes.electricalStorage, 0, 0],
             ent.CommonLabels.label_unique.value: ["iceStorage__B001", "shStorage__B001", "dhwStorage__B003",
                                                   "electricalStorage", 0, 0],
             ent.StorageLabels.initial_capacity.value: [0.25, 0.5, 0.3, 0.75, 0, 0],
             ent.StorageLabels.initial_temp.value: [1.5, "x", "x", "x", "x", "x"],
             }
        )}
        building_model_params = _pd.DataFrame(
            {ent.BuildingModelParameters.building_unique.value: ["Building_model__B001", "Building_model__B002"],
             ent.BuildingModelParameters.tWallInit: [10., 13.],
             ent.BuildingModelParameters.tIndoorInit: [22., 20.3],
             ent.BuildingModelParameters.tDistributionInit: [30., 29.],
             }
        )
        result, label_to_sheet = mpci.prep_mpc_inputs(nodal_data, building_model_params)

        errors = []
        try:
            self.assertDictEqual(result, {'dhwStorage__B003': {'initial capacity': 0.3},
                                          'electricalStorage': {'initial capacity': 0.75},
                                          'shStorage__B001': {'initial capacity': 0.5},
                                          'iceStorage__B001': {'initial capacity': 0.25, 'initial_temp': 1.5},
                                          'Building_model__B001': {'tDistributionInit': 30.0, 'tIndoorInit': 22.0,
                                                                   'tWallInit': 10.0},
                                          'Building_model__B002': {'tDistributionInit': 29.0, 'tIndoorInit': 20.3,
                                                                   'tWallInit': 13.0},
                                          }
                                 )
        except AssertionError as e:
            errors.append(e)

        try:
            self.assertDictEqual(label_to_sheet,
                                 {'Building_model__B001': ent.NodeKeysOptional.building_model_parameters,
                                  'Building_model__B002': ent.NodeKeysOptional.building_model_parameters,
                                  'dhwStorage__B003': ent.NodeKeys.storages,
                                  'electricalStorage': ent.NodeKeys.storages,
                                  'iceStorage__B001': ent.NodeKeys.storages,
                                  'shStorage__B001': ent.NodeKeys.storages
                                  })
        except AssertionError as e:
            errors.append(e)

        if errors:
            raise ExceptionGroup(f"Found {len(errors)} issues", errors)


STORAGES_SHEET_NAME = ent.NodeKeys.storages
BUILDING_MODEL_SHEET_NAME = ent.NodeKeysOptional.building_model_parameters


class TestMpcHandler(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_get_mpc_scenario_from_csv(self):
        """End2end test for this method.
        Includes assert of part of nodal_data state relevant for MPC.

        Maybe Flaky
        Seems to introduce the "building_model_parameter when run in the terminal
        """

        # Profile data is located relative to example scripts.
        # Therefore, we need to change directories.
        cur_dir = _os.getcwd()
        _os.chdir(xlsh.EXAMPLE_SCRIPT_DIR)
        input_folder_path = xlsh.ROOT_DATA_DIR / "CSVs" / "MPC_example_CSVs"
        mpc = mpci.MpcHandler(prediction_window_in_hours=24, time_step_in_minutes=60,
                              nr_of_buildings=1)
        mpc.set_full_time_period(2018, 1, 1, 2018, 1, 31)

        mpc.optimization_settings = en.OptimizationProperties(
            optimization_type="costs",
            merge_link_buses=False,
            merge_buses=None,
            merge_heat_source_sink=False,
            temperature_levels=False,
            cluster_size=None,
            dispatch_mode=True,
            include_carbon_benefits=False,
        )

        system_state = mpc.get_mpc_scenario_from_csv(input_folder_path)

        _os.chdir(cur_dir)

        errors = []
        try:
            self.assertDictEqual(system_state, {'electricalStorage__B001': {'initial capacity': 0},
                                                'shStorage__B001': {'initial capacity': 0}})
        except AssertionError as e:
            errors.append(e)

        try:
            assert mpc.nr_of_buildings == 1
        except AssertionError as e:
            errors.append(e)

        try:
            assert mpc.prediction_window_in_hours == 24
        except AssertionError as e:
            errors.append(e)

        try:
            assert mpc.time_step_in_minutes == 60
        except AssertionError as e:
            errors.append(e)

        try:
            assert all(mpc.nodal_data[ent.NodeKeys.storages][ent.StorageLabels.initial_capacity]) == 0
        except AssertionError as e:
            errors.append(e)

        if errors:
            raise ExceptionGroup(f"Found {len(errors)} issues", errors)

    def test_get_current_time_period(self):
        """Unit test"""
        current_time_period_start = _pd.DatetimeIndex(["2018-01-01"])[0]
        mpc = mpci.MpcHandler(prediction_window_in_hours=2, time_step_in_minutes=60,
                              nr_of_buildings=1)
        current_time_period = mpc.get_current_time_period(current_time_period_start)
        expected = _pd.DatetimeIndex(["2018-01-01 00:00:00", "2018-01-01 01:00:00", "2018-01-01 02:00:00"])
        _pd.testing.assert_index_equal(current_time_period, expected)

    def test_update_nodal_data(self):
        """Unit test"""
        current_system_state = {'electricalStorage__B001': {'initial capacity': 0.42},
                                'shStorage__B001': {'initial capacity': 0.66}}
        mpc = mpci.MpcHandler(prediction_window_in_hours=2, time_step_in_minutes=60,
                              nr_of_buildings=1)
        mpc.nodal_data = {STORAGES_SHEET_NAME: _pd.DataFrame([
            {ent.CommonLabels.label_unique: 'electricalStorage__B001', 'initial capacity': 0.},
            {ent.CommonLabels.label_unique: 'shStorage__B001', 'initial capacity': 0.},
        ])
        }
        mpc.label_to_sheet = {'electricalStorage__B001': STORAGES_SHEET_NAME,
                              'shStorage__B001': STORAGES_SHEET_NAME}

        current_nodal_data = mpc.update_nodal_data(current_system_state)

        df_expected = _pd.DataFrame(
            [{ent.CommonLabels.label_unique: 'electricalStorage__B001', 'initial capacity': 0.42},
             {ent.CommonLabels.label_unique: 'shStorage__B001', 'initial capacity': 0.66},
             ])
        _pd.testing.assert_frame_equal(current_nodal_data[STORAGES_SHEET_NAME], df_expected)

    def test_update_nodal_data_with_building(self):
        current_system_state = {'electricalStorage__B001': {'initial capacity': 0.42},
                                'shStorage__B001': {'initial capacity': 0.66},
                                'Building_model__B001': {'tDistributionInit': 25.1, 'tIndoorInit': 23.9,
                                                         'tWallInit': 11.3},
                                'Building_model__B002': {'tDistributionInit': 27.3, 'tIndoorInit': 21.5,
                                                         'tWallInit': 17.8}
                                }
        mpc = mpci.MpcHandler(prediction_window_in_hours=2, time_step_in_minutes=60,
                              nr_of_buildings=1)
        mpc.nodal_data = {
            STORAGES_SHEET_NAME: _pd.DataFrame([
                {ent.CommonLabels.label_unique: 'electricalStorage__B001', 'initial capacity': 0.},
                {ent.CommonLabels.label_unique: 'shStorage__B001', 'initial capacity': 0.},
            ]),
            ent.NodeKeysOptional.building_model_parameters: _pd.DataFrame([
                {ent.BuildingModelParameters.building_unique: 'Building_model__B001', 'tDistributionInit': 30.0,
                 'tIndoorInit': 22.0, 'tWallInit': 10.0},
                {ent.BuildingModelParameters.building_unique: 'Building_model__B002', 'tDistributionInit': 29.0,
                 'tIndoorInit': 20.3, 'tWallInit': 13.0},
            ])
        }
        mpc.label_to_sheet = {'electricalStorage__B001': STORAGES_SHEET_NAME,
                              'shStorage__B001': STORAGES_SHEET_NAME,
                              'Building_model__B001': BUILDING_MODEL_SHEET_NAME,
                              'Building_model__B002': BUILDING_MODEL_SHEET_NAME,
                              }

        current_nodal_data = mpc.update_nodal_data(current_system_state)

        errors = []

        df_expected = _pd.DataFrame(
            [{ent.CommonLabels.label_unique: 'electricalStorage__B001', 'initial capacity': 0.42},
             {ent.CommonLabels.label_unique: 'shStorage__B001', 'initial capacity': 0.66},
             ])
        try:
            _pd.testing.assert_frame_equal(current_nodal_data[STORAGES_SHEET_NAME], df_expected)
        except AssertionError as e:
            errors.append(e)

        df_expected = _pd.DataFrame([
            {ent.BuildingModelParameters.building_unique: 'Building_model__B001', 'tDistributionInit': 25.1,
             'tIndoorInit': 23.9, 'tWallInit': 11.3},
            {ent.BuildingModelParameters.building_unique: 'Building_model__B002', 'tDistributionInit': 27.3,
             'tIndoorInit': 21.5, 'tWallInit': 17.8},
        ])

        try:
            _pd.testing.assert_frame_equal(current_nodal_data[BUILDING_MODEL_SHEET_NAME], df_expected)
        except AssertionError as e:
            errors.append(e)

        if errors:
            raise ExceptionGroup(f"Found {len(errors)} issues", errors)

    def test_get_flow_label_to_sheet(self):
        results = {
            "a": _pd.DataFrame(columns=["1", "2", "3"]),
            "b": _pd.DataFrame(columns=["4", "5", "6"]),
            "c": _pd.DataFrame(columns=["7", "8", "9"]),
        }
        expected_label_to_sheet = {
            "1": "a",
            "2": "a",
            "3": "a",
            "4": "b",
            "5": "b",
            "6": "b",
            "7": "c",
            "8": "c",
            "9": "c",
        }
        mpc = mpci.MpcHandler(prediction_window_in_hours=2, time_step_in_minutes=60,
                              nr_of_buildings=1)
        flow_label_to_sheet = mpc.get_flow_label_to_sheet(results)

        self.assertDictEqual(flow_label_to_sheet, expected_label_to_sheet)

    def test_rename_oemof_labels(self):
        """Unit test
        Describes the expected labels from oemof and the expected behavior for the renaming.

        The dictionary includes the names that should be renamed.
        The other names added to the input_names should be ignored by the renaming, which is why they are not
        added to the initial dictionary.
        """
        input_and_expected_names = {
            "(('pv__Building1', 'electricityProdBus__Building1'), 'flow')": "pv__B001__To__electricityProdBus__B001",
            "(('pv__Building567', 'electricityProdBus__Building567'), 'flow')":
                "pv__B567__To__electricityProdBus__B567",
            "(('electricityBus__Building1', 'excesselectricityBus__Building1'), 'flow')":
                "electricityBus__B001__To__excesselectricityBus__B001",
            "(('electricityProdBus__Building1', 'electricalStorage__Building1'), 'flow')":
                "electricityProdBus__B001__To__electricalStorage__B001",
            "(('electricalStorage__Building1', 'electricityBus__Building1'), 'flow')":
                "electricalStorage__B001__To__electricityBus__B001",
            "(('electricityBus__Building1', 'producedElectricity__Building1'), 'flow')":
                "electricityBus__B001__To__producedElectricity__B001",
            "(('electricityResource__Building1', 'gridBus__Building1'), 'flow')":
                "electricityResource__B001__To__gridBus__B001",
            "(('electricityInBus__Building1', 'HP__Building1'), 'flow')": "electricityInBus__B001__To__HP__B001",
            "(('HP__Building1', 'shSourceBus__Building1'), 'flow')": "HP__B001__To__shSourceBus__B001",
            "(('shSourceBus__Building1', 'shStorage__Building1'), 'flow')": "shSourceBus__B001__To__shStorage__B001",
            "(('shSourceBus__Building1', 'shSource__Building1'), 'flow')": "shSourceBus__B001__To__shSource__B001",
            "(('shStorage__Building1', 'spaceHeatingBus__Building1'), 'flow')":
                "shStorage__B001__To__spaceHeatingBus__B001",
            "(('spaceHeatingBus__Building1', 'spaceHeating__Building1'), 'flow')":
                "spaceHeatingBus__B001__To__spaceHeating__B001",
        }
        input_names = list(input_and_expected_names.keys())
        input_names += ["storage_content",  # TODO: check whether this should become "sh_storage_content"
                        "some_flow_to_be_ignored",
                        0,  # This occurs in the capacity sheets and should be fixed at a different level.
                        42,
                        ]

        rename_dict = mpci.MpcHandler.rename_oemof_labels(input_names)

        self.assertDictEqual(rename_dict, input_and_expected_names)

    def test_get_desired_energy_flows(self):
        """Unit test.
        This method should both rename the columns and collect the desired ones into a single data structure.
        """
        desired_flows_with_new_names = {
            "pv__B001__To__electricityProdBus__B001": "el_pv_produced",
            "electricityBus__B001__To__excesselectricityBus__B001": "el_to_grid",
            "electricityProdBus__B001__To__electricalStorage__B001": "el_pv_to_battery",
            "electricalStorage__B001__To__electricityBus__B001": "el_battery_discharge",
            "electricityBus__B001__To__producedElectricity__B001": "el_produced",
            "electricityResource__B001__To__gridBus__B001": "el_from_grid",
            "electricityInBus__B001__To__HP__B001": "HP_el_in",
            "HP__B001__To__shSourceBus__B001": "HP_heat_out",
            "shSourceBus__B001__To__shStorage__B001": "HP_to_TES",
            "shSourceBus__B001__To__shSource__B001": "HP_to_demand",
            "shStorage__B001__To__spaceHeatingBus__B001": "TES_to_demand",
            "spaceHeatingBus__B001__To__spaceHeating__B001": "sh_delivered",
        }
        expected_energy_flows = _pd.DataFrame(
            {
                "el_pv_produced": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.108543350181667, 8.272728965148124,
                                   50.89640662861397, 17.94914364741515, 7.54176286459802, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                   0.0, 0.0, 0.0, 0.0, 0.0],
                "el_to_grid": [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 5.586883168290823, 46.5148636963539,
                               4.562246680843671, 6.47183429359802, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                               15.02745821417941, 0.0, 0.0, 0.0],
                "el_pv_to_battery": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ],
                "el_battery_discharge": [0.554028571, 0.469628571, 7.245377247649912, 0.549585714, 0.6267,
                                         3.771141997591101, 6.202756082034924, 5.529049211325205, 5.907222702238744,
                                         3.144218023214221, 0.0, 0.0, 0.0, 0.0, 2.825040474208677, 3.212498410367769,
                                         1.398457143, 1.839342857, 2.005842857, 8.649585714, 12.11352857,
                                         25.47546728336945, 1.689628571, 0.813671429, 0.577228571],
                "el_produced": [0.554028571, 0.469628571, 7.245377247649912, 0.549585714, 0.6267, 3.771141997591101,
                                6.202756082034924, 5.529049211325205, 5.907222702238744, 4.252761373395888,
                                2.6858457968573, 4.381542932260076, 13.38689696657148, 1.069928571, 2.825040474208677,
                                3.212498410367769, 1.398457143, 1.839342857, 2.005842857, 8.649585714, 12.11352857,
                                10.44800906919004, 1.689628571, 0.813671429, 0.577228571],
                "el_from_grid": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ],
                "HP_el_in": [0.0, 0.0, 6.788420104649911, 0.0, 0.0, 3.160941997591101, 2.948527511034924,
                             2.792563497325205, 2.451579845238745, 1.688489944395888, 1.3952600828573,
                             0.08261436126007536, 11.45105410957148, 0.0, 1.766354760208677, 1.492741267367769, 0.0,
                             0.0, 0.0, 0.0, 0.0, 6.288094783190043, 0.0, 0.0, 0.0],
                "HP_heat_out": [0.0, 0.0, 28.77093082197741, 0.0, 0.0, 13.39682019424264, 12.49655733424264,
                                11.86780923424264, 10.47549639424264, 7.293487614242636, 6.043188243242636,
                                0.3715253452426365, 52.325, 0.0, 8.007204924272527, 6.730931781940396, 0.0, 0.0, 0.0,
                                0.0, 0.0, 28.0525510618056, 0.0, 0.0, 0.0],
                "HP_to_TES": [0.0, 0.0, 15.59785494197741, 0.0, 0.0, 0.0805135742426365, 0.0805135742426365,
                              0.0805135742426365, 0.0805135742426365, 0.0805135742426365, 0.0805135742426365,
                              0.0805135742426365, 52.325, 0.0, 3.353566187272527, 0.1341832079403957, 0.0, 0.0, 0.0,
                              0.0, 0.0, 20.6793614738056, 0.0, 0.0, 0.0],
                "HP_to_demand": [0.0, 0.0, 13.17307588, 0.0, 0.0, 13.31630662, 12.41604376, 11.78729566, 10.39498282,
                                 7.21297404, 5.962674669, 0.291011771, 0.0, 0.0, 4.653638737, 6.596748574, 0.0, 0.0,
                                 0.0, 0.0, 0.0, 7.373189588, 0.0, 0.0, 0.0],
                "TES_to_demand": [13.14460485, 13.16625478, 0.0, 13.17181806, 13.27046787, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                  0.0, 0.0, 3.008074284, 0.0, 0.0, 7.847446237, 8.90450432, 9.358278467, 7.240053475,
                                  6.579958661, 0.0, 9.326260232, 10.91023682, 11.85101617],
                "sh_delivered": [13.14460485, 13.16625478, 13.17307588, 13.17181806, 13.27046787, 13.31630662,
                                 12.41604376, 11.78729566, 10.39498282, 7.21297404, 5.962674669, 0.291011771, 0.0,
                                 3.008074284, 4.653638737, 6.596748574, 7.847446237, 8.90450432, 9.358278467,
                                 7.240053475, 6.579958661, 7.373189588, 9.326260232, 10.91023682, 11.85101617],
            },
            index=_pd.date_range("2018-01-01 00:00:00", "2018-01-02 00:00:00", freq="60min"),
        )
        expected_energy_flows.index.freq = None

        _input_data_path = _pl.Path(
            __file__).parent / "expected_files" / "results_MPC_example_2018_01_01__00_00_00.xlsx"
        data = _pd.ExcelFile(str(_input_data_path))
        results = {}
        for sheet in data.sheet_names:
            if any(sheet.startswith(s) for s in ["cap", "costs_", "env_"]):
                results[sheet] = data.parse(sheet)
            else:
                results[sheet] = data.parse(sheet, index_col=0)
        data.close()

        mpc = mpci.MpcHandler(prediction_window_in_hours=2, time_step_in_minutes=60,
                              nr_of_buildings=1)
        energy_flows = mpc.get_desired_energy_flows(results, desired_flows_with_new_names)

        _pd.testing.assert_frame_equal(energy_flows, expected_energy_flows)



import sys
from random import random
from matplotlib import colors
import numpy as np
import os
import pandas as pd
import pathlib as _pl
import plotly.graph_objects as go

from optihood.plot_functions import getData
from optihood.labelDict import labelDictGenerator, positionDictGenerator


def addCapacities(nodes, dataDict, buildings, UseLabelDict, labelDict, mergedLinks):
    capacities = ["sufficient"] * len(nodes)
    for i in buildings:
        capTransformers=dataDict["capTransformers__Building"+str(i)]
        capStorages=dataDict["capStorages__Building"+str(i)]
        for j, k in capStorages.iterrows():
            if k[0] < 0.1: # if the installed capacity is 0 then skip (sometimes as an error very low capacites are selected. To handle this k<0.01kW is set as the condition for comparison)
                continue
            index=nodes.index(labelDict[j])
            #nodes[index]=nodes[index]+" "+str(round(k[0],2))+" kWh"
            if "Bat" in labelDict[j]:
                capacities[index] = str(round(k[0], 1)) + " kWh"
            else:
                capacities[index] = str(round(k[0], 1)) + " L"
        for j, k in capTransformers.iterrows():
            if k[0] < 0.001:     # if the installed capacity is 0 then skip (sometimes as an error very low capacites are selected. To handle this k<0.001kW is set as the condition for comparison
                continue
            jComponents=j.split("'")
            if 'Bus' in jComponents[1]:
                j =jComponents[3]
            else:
                j=jComponents[1]

            index=nodes.index(labelDict[j])

            #nodes[index]=nodes[index]+" "+str(k[0])+" kW"
            if round(k[0],1) == 0:
                capacities[index] = str(round(k[0], 2)) + " kW"
            else:
                capacities[index] = str(round(k[0], 1)) + " kW"
    return capacities


def readResults(fileName, buildings, ColorDict, UseLabelDict, labelDict, positionDict, labels, mergedLinks):
    dataDict = getData(fileName)
    keys=dataDict.keys()
    nodes, sources, targets, values,x,y = createSankeyData(dataDict, keys, UseLabelDict, labelDict, positionDict, buildings, mergedLinks)
    capacities = addCapacities(nodes, dataDict, buildings, UseLabelDict, labelDict, mergedLinks)
    nodesColors = pd.Series(createColorList(nodes, ColorDict, labels))
    linksColors = nodesColors[sources]
    linksColors = np.where(nodesColors[targets] == ColorDict["dhw"], ColorDict["dhw"], linksColors)
    linksColors = np.where(nodesColors[targets] == ColorDict["sh"], ColorDict["sh"], linksColors)
    linksColors = np.where(nodesColors[targets] == ColorDict["elec"], ColorDict["elec"], linksColors)

    data = [go.Sankey(
        arrangement="snap",
        valuesuffix="kWh",
        node={
            "pad":25,
            "thickness":15,
            "line":dict(color="black", width=0.5),
            "label":nodes,#+str(values),
            "color":nodesColors.tolist(),
            #"groups":[linkGroup],
            "customdata": capacities,
            "hovertemplate":  '%{label} has %{customdata} capacity installed',
            "x":x,
            "y":y,
            },
        link={
            "source":sources,
            "target":targets,
            "value":values,
            "color":linksColors.tolist(),
            },
        )]
    return data


def createSankeyData(dataDict, keys, UseLabelDict, labelDict, PositionDict, buildings=[], mergedLinks=False):
    sources = [] #contains index of node
    targets = [] #contains index of node
    values = []
    nodes = []
    # (x,y) is the position of the node
    x=[] #equivalent in dimension to nodes
    y=[] #equivalent in dimension to nodes
    linkGroup=[]
    mergedComponents = ["electricityBus", "electricityInBus", "domesticHotWaterBus", "dhwDemandBus", "spaceHeatingBus", "shDemandBus", "excesselectricityBus", "producedElectricity", "spaceHeating", "domesticHotWater"]

    for key in keys:
        df = dataDict[key]
        dfKeys = df.keys()
        if "dhwStorageBus" in key and not mergedComponents:
            continue
        if all([str(i) not in key for i in buildings]) and key not in mergedComponents:
            continue
        for dfKey in dfKeys:
            if isinstance(dfKey, int) or "storage_content" in dfKey:
                continue

            dfKeySplit = dfKey.split("'")
            sourceNodeName=dfKeySplit[1]
            targetNodeName =dfKeySplit[3]

            if mergedLinks:
                # for the sake for representation the merged buses (if present) are added to Building 1
                if sourceNodeName in mergedComponents:
                    sourceNodeName = sourceNodeName + '__Building1'
                if targetNodeName in mergedComponents:
                    targetNodeName = targetNodeName + '__Building1'


            if sourceNodeName in labelDict:
                sourceNodeName = labelDict[sourceNodeName]
            if targetNodeName in labelDict:
                targetNodeName = labelDict[targetNodeName]
            if sourceNodeName == targetNodeName:
                continue
            if "exSolar" in targetNodeName:
                continue

            if "Resource" not in sourceNodeName:
                dfKeyValues = df[dfKey].values
                value = sum(dfKeyValues)
                if value < 0.001:
                    continue
                values.append(value)
                if sourceNodeName not in nodes:
                    nodes.append(sourceNodeName)
                    for posKey in PositionDict.keys():
                        if posKey in sourceNodeName and posKey[0:2] == sourceNodeName[0:2]: #second part of the term added for CHP and HP
                            x.append(PositionDict[posKey][0])
                            if labelDict["electricityLink"] in sourceNodeName:
                                y.append((0.5-(PositionDict[posKey][1]))/len(buildings))
                            elif labelDict["shLink"] in sourceNodeName:
                                y.append((0.5 - (PositionDict[posKey][1])) / len(buildings))
                            elif labelDict["dhwLink"] in sourceNodeName:
                                y.append((0.5 - (PositionDict[posKey][1])) / len(buildings))
                            elif ("grid" in sourceNodeName or "Grid" in sourceNodeName) and mergedLinks:
                                buildingNumber = 1
                                temp = (PositionDict[posKey][1]) / len(buildings) + buildingNumber / len(buildings)
                                y.append(temp)
                            else:
                                buildingNumber=buildings.index(int(sourceNodeName.split('_')[-1][1:]))
                                temp = (PositionDict[posKey][1]) / len(buildings) + buildingNumber / len(buildings)
                                y.append(temp)
                sources.append(nodes.index(sourceNodeName))

                if targetNodeName not in nodes:
                    nodes.append(targetNodeName)
                    for posKey in PositionDict.keys():
                        if posKey in targetNodeName and posKey[0:2] == targetNodeName[0:2]:
                            x.append(PositionDict[posKey][0])
                            if labelDict["electricityLink"] in targetNodeName:
                                y.append((0.5-(PositionDict[posKey][1]))/len(buildings))
                            elif labelDict["shLink"] in targetNodeName:
                                y.append((0.5 - (PositionDict[posKey][1])) / len(buildings))
                            elif labelDict["dhwLink"] in targetNodeName:
                                y.append((0.5 - (PositionDict[posKey][1])) / len(buildings))
                            elif ("grid" in sourceNodeName or "Grid" in sourceNodeName) and mergedLinks:
                                buildingNumber = 1
                                temp = (PositionDict[posKey][1]) / len(buildings) + buildingNumber / len(buildings)
                                y.append(temp)
                            else:
                                buildingNumber=buildings.index(int(targetNodeName.split('_')[-1][1:]))
                                temp = (PositionDict[posKey][1]) / len(buildings) + (buildingNumber) / len(buildings)
                                y.append(temp)
                targets.append(nodes.index(targetNodeName))
    return nodes, sources, targets, values, x, y


def createColorList(inputList, ColorDict, labels):
    colorsList=[]
    for n in inputList:
        if (labels!='default' and labels["naturalGas"] in n) or (labels=='default' and "natGas" in n):           # Check for whether labels of a specific type are defined or not should be added here
            color = ColorDict["gas"]
        elif (labels!='default' and labels["qSource"] in n) or (labels=='default' and "qSource" in n):
            color = ColorDict["qs"]
        elif (labels!='default' and (labels["excessSh"] in n or labels["prodSH"] in n or labels["shBus"] in n or labels["StorageSh"] in n or labels["DemandSh"] in n)) or (labels=='default' and any(v in n for v in ["shLink", "prodSH", "shBus", "shStor", "Q_sh", "exSh", "usedSH"])):
            color = ColorDict["sh"]
        elif (labels!='default' and (labels["solarCollector"] in n or labels["excessSolarCollector"] in n or labels["StorageDhw"] in n or labels["dhwBus"] in n or labels["DemandDhw"] in n)) or (labels=='default' and any(v in n for v in ["dhwLink", "solar", "exSolar", "dhwStor", "dhwBus", "Q_dhw", "prodDHW"])):
            color = ColorDict["dhw"]
        elif (labels!='default' and (labels["elBus"] in n or labels["grid"] in n or labels["pv"] in n or labels["prodEl"] in n or labels["localEl"] in n or labels["StorageEl"] in n or labels["excessEl"] in n or labels["DemandEl"] in n or labels["DemandMob"] in n))  or (labels=='default' and any(v in n for v in ["excesselectricityProdBus", "eMobilityDemand", "elLink", "grid", "pv", "prodEl", "localEl", "Bat", "exEl", "usedEl", "Q_el", "Q_mob"])):
            color = ColorDict["elec"]
        elif (labels!='default' and labels["hSB"] in n) or (labels=='default' and "hSB" in n):
            color = ColorDict["hs"]
        else:
            color = ColorDict["other"]
        colorsList.append(color)
    return colorsList


def displaySankey(fileName, UseLabelDict, labelDict, positionDict, labels, buildings, mergedLinks, hideBuildingNumber):
    OPACITY = 0.6
    ColorDict = {"elec": 'rgba' + str(colors.to_rgba("skyblue", OPACITY)),
                 "gas": 'rgba' + str(colors.to_rgba("darkgray", OPACITY)),
                 "dhw": 'rgba' + str(colors.to_rgba("red", OPACITY)),
                 "sh": 'rgba' + str(colors.to_rgba("magenta", OPACITY)),
                 "other": 'rgba' + str(colors.to_rgba("deeppink", OPACITY)),
                 "hs": 'rgba' + str(colors.to_rgba("yellow", OPACITY)),
                 "qs": 'rgba' + str(colors.to_rgba("green", OPACITY))
                 }
    data = readResults(fileName, buildings, ColorDict, UseLabelDict, labelDict, positionDict, labels, mergedLinks)

    file_path = _pl.Path(fileName)
    title_name_for_file = file_path.stem + file_path.suffix

    node = data[0]['node']
    link = data[0]['link']
    if hideBuildingNumber == True:
        node['label'] = tuple(s if not "_B" in s else s.split("_")[0] for s in node['label'])
    fig = go.Figure(go.Sankey(arrangement = "perpendicular",
                              link=link,
                              node=node
                              )) #snap, perpendicular,freeform, fixed
    fig.update_layout(
        title=title_name_for_file +" for buildings " + str(buildings),
        font=dict(size=10, color='black'),
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
    )
    fig.add_hline(y=0, line_color='rgba(0,0,0,0)')
    if len(buildings)%2==0:
        fig.add_hline(y=0.5, line_dash="dash")
    if len(buildings)%3==0:
        fig.add_hline(y=0.33, line_dash="dash")
        fig.add_hline(y=0.66, line_dash="dash")
    if len(buildings)%4==0:
        fig.add_hline(y=0.25, line_dash="dash")
        fig.add_hline(y=0.75, line_dash="dash")
    fig.add_hline(y=1.0, line_color='rgba(0,0,0,0)')
    fig.update_xaxes(visible=False)
    fig.update_yaxes(visible=False)
    return fig


def plot(excelFileName, outputFileName, numberOfBuildings, UseLabelDict, labels, optimType, mergedLinks=False, hideBuildingNumber=False, show_figs=True):
    BUILDINGSLIST = list(range(1, numberOfBuildings + 1))
    labelDict = labelDictGenerator(numberOfBuildings, labels, optimType, mergedLinks)
    positionDict = positionDictGenerator(labels, optimType, mergedLinks)
    fig = displaySankey(excelFileName, UseLabelDict, labelDict, positionDict, labels, BUILDINGSLIST, mergedLinks, hideBuildingNumber)

    fig.write_html(outputFileName, div_id="0fbea2e5-8f67-4d21-8fe7-897e199ac035")  # this is hardcoded to make the htmls reproducible
    if show_figs:
        fig.show()


if __name__ == "__main__":
    optMode = "group"  # parameter defining whether the results file corresponds to "indiv" or "group" optimization
    numberOfBuildings = 4
    plotOptim = 3  # defines the number of the optimization to plot
    UseLabelDict = True
    excelFileName = os.path.join(r"..\data\Results", f"results{numberOfBuildings}_{plotOptim}_{optMode}.xlsx")
    outputFileName = os.path.join(r"..\data\figures", f"Sankey_{plotOptim}.html")
    plot(excelFileName, outputFileName, numberOfBuildings, UseLabelDict, labels='default', optimType=optMode)
